from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal, InvalidOperation
from logging import getLogger
from string import Formatter
from typing import Any, ClassVar, Generic, TypeVar, overload

from whenever import Date, Instant, OffsetDateTime, PlainDateTime, Time, ZonedDateTime

from hassette.exceptions import UnableToConvertValueError
from hassette.utils.date_utils import convert_datetime_str_to_system_tz

R = TypeVar("R")
T = TypeVar("T")

ALLOWED_FORMAT_FIELDS = {"value", "from_type", "to_type"}
"""Allowed fields for formatting error messages in type converters."""

LOGGER = getLogger(__name__)


def get_format_fields(string_value: str) -> list[str]:
    """Get the format fields from a string.

    Args:
        string_value: The string to parse.

    Returns:
        A list of format fields (e.g. ["field1", "field2"]).

    Example:
        >>> get_format_fields("Hello, {name}!")
        ['name']
    """
    return [fn for _, fn, _, _ in Formatter().parse(string_value) if fn is not None]


@dataclass
class TypeConverterEntry(Generic[T, R]):
    """Represents a type conversion function and its associated metadata."""

    func: Callable[[T], R]
    from_type: type[T]
    to_type: type[R]
    error_types: tuple[type[BaseException], ...] = (ValueError,)
    error_message: str | None = None


@overload
def register_type_converter_fn(fn: Callable[[T], R]) -> Callable[[T], R]: ...


@overload
def register_type_converter_fn(
    fn: None = None, *, error_message: str | None = None, error_types: tuple[type[BaseException], ...] = (ValueError,)
) -> Callable[[Callable[[T], R]], Callable[[T], R]]: ...


def register_type_converter_fn(
    fn: Callable[[T], R] | None = None,
    *,
    error_message: str | None = None,
    error_types: tuple[type[BaseException], ...] = (ValueError,),
):
    """Register a type conversion function with the TypeRegistry.

    Can be used as:

        @register_type_converter
        def convert_x(value: T) -> R: ...

    or:

        @register_type_converter(error_message="failed to convert X")
        def convert_x(value: T) -> R: ...
    """
    if error_message is not None:
        fields = get_format_fields(error_message)
        invalid_fields = set(fields) - ALLOWED_FORMAT_FIELDS
        if invalid_fields:
            raise ValueError(f"Invalid format fields in error_message: {invalid_fields}")

    def decorator(func: Callable[[T], R]) -> Callable[[T], R]:
        from_type = func.__annotations__["value"]
        to_type = func.__annotations__["return"]
        TypeRegistry.register(
            TypeConverterEntry(
                func=func, from_type=from_type, to_type=to_type, error_message=error_message, error_types=error_types
            )
        )
        return func

    # Used as bare @register_type_converter
    if fn is not None:
        return decorator(fn)

    # Used as @register_type_converter(...)
    return decorator


def register_simple_type_converter(
    from_type: type[T],
    to_type: type[R],
    fn: Callable[[T], R] | None = None,
    error_message: str | None = None,
    error_types: tuple[type[BaseException], ...] = (ValueError,),
):
    """Register a simple type conversion function from a non-user defined function, such as a constructor.

    Args:
        from_type: The source type to convert from.
        to_type: The target type to convert to.
        fn: The function to use for conversion. If None, the target type constructor is used.
        error_message: Optional custom error message if conversion fails.
        error_types: Tuple of exception types to catch and wrap in UnableToConvertValueError.

    Example:
        register_simple_type_converter(int, float, error_message="Failed to convert int to float")
        register_simple_type_converter(ZonedDateTime, str, fn=ZonedDateTime.format_iso)
    """
    if error_message is not None:
        fields = get_format_fields(error_message)
        invalid_fields = set(fields) - ALLOWED_FORMAT_FIELDS
        if invalid_fields:
            raise ValueError(f"Invalid format fields in error_message: {invalid_fields}")

    fn = fn or (lambda x: to_type(x))  # pyright: ignore[reportCallIssue]

    TypeRegistry.register(
        TypeConverterEntry(
            func=fn,
            from_type=from_type,
            to_type=to_type,
            error_message=error_message,
            error_types=error_types,
        )
    )


class TypeRegistry:
    """Registry for converting between types, used by State models and the Dependency Injection system."""

    conversion_map: ClassVar[dict[tuple[type[Any], type[Any]], TypeConverterEntry[Any, Any]]] = {}

    @classmethod
    def register(cls, type_converter: TypeConverterEntry[Any, Any]) -> None:
        """Register a type converter in the registry."""
        from_type = type_converter.from_type
        to_type = type_converter.to_type
        key = (from_type, to_type)
        if key in cls.conversion_map:
            LOGGER.warning("Overwriting existing conversion from %s to %s", from_type.__name__, to_type.__name__)
        cls.conversion_map[key] = type_converter

    def convert(self, value: Any, to_type: type[Any] | tuple[type[Any], ...]) -> Any:
        """Convert a StateValue to a target Python type.

        Args:
            value: The StateValue instance to convert.
            to_type: The target Python type.

        Returns:
            The converted value.
        """

        # handle tuple
        if isinstance(to_type, tuple):
            if isinstance(value, to_type):
                LOGGER.debug("Value %r is already of type %s, no conversion needed", value, type(value).__name__)
                return value

            for tt in to_type:
                with suppress(UnableToConvertValueError):
                    return self.convert(value, tt)
            raise UnableToConvertValueError(f"Unable to convert {value!r} to any of the types {to_type}")

        # handle single type

        from_type = type(value)
        key = (from_type, to_type)

        # handle Any type
        if to_type is type(Any):
            return value

        # handle exact type match
        if to_type is from_type:
            return value

        # handle None value
        if value is None:
            return value

        if to_type is type(None) and value is not None:
            LOGGER.debug("Not attempting to convert %r to NoneType", value)
            raise UnableToConvertValueError(f"Cannot convert {value!r} to NoneType")

        # if we don't have this in our map, attempt to just convert using it as a constructor
        if key not in self.conversion_map and to_type is not type(None):
            try:
                new_value = to_type(value)
                LOGGER.debug(
                    "Converted %r (%s) to %r (%s) using constructor",
                    value,
                    type(value).__name__,
                    new_value,
                    to_type.__name__,
                )
                return new_value
            except Exception as e:
                raise UnableToConvertValueError(f"Unable to convert {value!r} to {to_type}") from e

        fn = self.conversion_map[key]

        try:
            new_value = fn.func(value)
            LOGGER.debug(
                "Converted %r (%s) to %r (%s) using registered converter %s",
                value,
                type(value).__name__,
                new_value,
                to_type.__name__,
                fn.func.__name__,
            )
            return new_value
        except fn.error_types as e:
            default_err_msg = f"Error converting {value!r} ({type(value).__name__}) to {to_type.__name__}"
            err_msg = fn.error_message or default_err_msg
            if get_format_fields(err_msg):
                err_msg = err_msg.format(value=value, from_type=from_type, to_type=to_type)

            LOGGER.debug("Error converting %r (%s) to %s: %s", value, type(value).__name__, to_type.__name__, err_msg)

            raise UnableToConvertValueError(err_msg) from e
        except Exception as e:
            raise RuntimeError(f"Error converting {value!r} ({type(value).__name__}) to {to_type.__name__}") from e

    def list_conversions(self) -> list[tuple[type, type, TypeConverterEntry]]:
        """List all registered type conversions.

        Returns a sorted list of all registered type conversions with their metadata.
        Useful for debugging and inspection of available converters.

        Returns:
            List of (from_type, to_type, entry) tuples sorted by from_type name then to_type name.

        Example:
            ```python
            from hassette import TYPE_REGISTRY

            # List all conversions
            conversions = TYPE_REGISTRY.list_conversions()
            for from_type, to_type, entry in conversions:
                print(f"{from_type.__name__} → {to_type.__name__}: {entry.description}")
            ```
        """
        items = []
        for (from_type, to_type), entry in self.conversion_map.items():
            items.append((from_type, to_type, entry))

        # Sort by from_type name, then to_type name
        items.sort(key=lambda x: (x[0].__name__, x[1].__name__))
        return items


## Value Converters ##

# stdlib classes
register_simple_type_converter(Decimal, float)
register_simple_type_converter(Decimal, int)
register_simple_type_converter(float, Decimal, error_types=(ValueError, InvalidOperation))
register_simple_type_converter(float, int)
register_simple_type_converter(float, str)
register_simple_type_converter(int, float)
register_simple_type_converter(bool, str)
register_simple_type_converter(int, str)
register_simple_type_converter(str, Decimal, error_types=(ValueError, InvalidOperation))
register_simple_type_converter(str, float)
register_simple_type_converter(str, int)


# non-stdlib classes
register_simple_type_converter(str, Date, fn=Date.parse_iso)
register_simple_type_converter(str, Time, fn=Time.parse_iso)
register_simple_type_converter(str, OffsetDateTime, fn=OffsetDateTime.parse_iso)
register_simple_type_converter(str, PlainDateTime, fn=PlainDateTime.parse_iso)
register_simple_type_converter(Time, time, fn=Time.py_time)
register_simple_type_converter(Time, str, fn=Time.format_iso)
register_simple_type_converter(ZonedDateTime, Instant, fn=ZonedDateTime.to_instant)
register_simple_type_converter(ZonedDateTime, PlainDateTime, fn=ZonedDateTime.to_plain)
register_simple_type_converter(ZonedDateTime, str, fn=ZonedDateTime.format_iso)

# more complex converters


@register_type_converter_fn(error_message="String must be a datetime-like value, got {from_type}")
def from_string_to_zoned_date_time(value: str) -> ZonedDateTime:
    with suppress(ValueError):
        return convert_datetime_str_to_system_tz(value)
    with suppress(ValueError):
        return PlainDateTime.parse_iso(value).assume_system_tz()
    with suppress(ValueError):
        return Date.parse_iso(value).at(Time(0, 0, 0, nanosecond=0)).assume_system_tz()
    raise ValueError


@register_type_converter_fn(error_message="String must be a time-like value, got {from_type}")
def from_string_to_stdlib_time(value: str) -> time:
    return Time.parse_iso(value).py_time()


@register_type_converter_fn(error_message="String must be a date-like value, got {from_type}")
def from_string_to_stdlib_date(value: str) -> date:
    return Date.parse_iso(value).py_date()


@register_type_converter_fn(error_message="String must be a datetime-like value, got {from_type}")
def from_string_to_stdlib_datetime(value: str) -> datetime:
    return from_string_to_zoned_date_time(value).py_datetime()


@register_type_converter_fn(error_message="String must be a boolean-like value, got {from_type}")
def from_string_to_bool(value: str) -> bool:
    lower_val = value.lower()
    match lower_val:
        case "on" | "true" | "yes" | "1":
            return True
        case "off" | "false" | "no" | "0":
            return False
        case _:
            raise ValueError



from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, ClassVar, Literal, get_args, get_origin

from hassette.utils.type_utils import (
    NoneType,
    is_optional,
    is_union,
    normalize_annotation,
    normalize_for_isinstance,
    safe_isinstance,
)


@dataclass(frozen=True)
class TypeMatcherEntry:
    """How to deeply match a runtime value against a parameterized origin[T,...]."""

    origin: Any
    match_fn: Callable[["TypeMatcher", Any, Any], bool]
    name: str | None = None


class TypeMatcherRegistry:
    """Registry mapping generic origins (list/dict/tuple/...) to deep match functions."""

    map: ClassVar[dict[Any, TypeMatcherEntry]] = {}

    @classmethod
    def register(cls, entry: TypeMatcherEntry) -> None:
        cls.map[entry.origin] = entry

    @classmethod
    def get(cls, origin: Any) -> TypeMatcherEntry | None:
        return cls.map.get(origin)


def register_type_matcher(*origins: Any):
    """Decorator to register deep-matching for one or more generic origins."""

    def deco(fn: Callable[["TypeMatcher", Any, Any], bool]) -> Callable[["TypeMatcher", Any, Any], bool]:
        for origin in origins:
            TypeMatcherRegistry.register(
                TypeMatcherEntry(origin=origin, match_fn=fn, name=getattr(fn, "__name__", None))
            )
        return fn

    return deco


class TypeMatcher:
    """Runtime matcher for checking if values already satisfy nested type annotations."""

    def matches(self, value: Any, tp: Any) -> bool:
        tp = normalize_annotation(tp)

        # Any matches everything
        if tp is Any:
            return True

        # Check optional prior to union to short-circuit
        if is_optional(tp):
            return value is None or any(self.matches(value, arg) for arg in get_args(tp) if arg is not NoneType)

        # Union: only return "True" if it matches the first type
        # otherwise we risk saying "sure, '6' matches [int | str]" when the caller really wanted an int
        if is_union(tp):
            args = get_args(tp)
            if not args:
                return False
            return self.matches(value, args[0])

        # Literal[...] is *very* useful for DI-style validation
        origin = get_origin(tp)
        if origin is Literal:
            return value in get_args(tp)

        # Non-parameterized type: use normalize_for_isinstance
        if origin is None:
            with suppress(TypeError):
                norm = normalize_for_isinstance(tp)
                return isinstance(value, norm)
            return False

        # Parameterized type: must satisfy outer container/protocol first
        with suppress(TypeError):
            if not isinstance(value, origin):
                return False
        # If origin can't be used with isinstance, treat as non-match (forces conversion).
        # If you'd rather be permissive, return True instead.
        if not safe_isinstance(value, origin):
            return False

        # Deep match for known container origins
        entry = TypeMatcherRegistry.get(origin)
        if entry is not None:
            return entry.match_fn(self, value, tp)

        # Unknown generic: outer match is the best we can do
        return True


@register_type_matcher(list, set, frozenset)
def match_homogeneous_iterable(m: TypeMatcher, value: Any, tp: Any) -> bool:
    args = get_args(tp)
    if not args:
        return True
    (elem_tp,) = args
    return all(m.matches(v, elem_tp) for v in value)


@register_type_matcher(dict)
def match_dict(m: TypeMatcher, value: Any, tp: Any) -> bool:
    args = get_args(tp)
    if len(args) != 2:
        return True
    key_tp, val_tp = args
    return all(m.matches(k, key_tp) and m.matches(v, val_tp) for k, v in value.items())


@register_type_matcher(tuple)
def match_tuple(m: TypeMatcher, value: Any, tp: Any) -> bool:
    args = get_args(tp)
    if not args:
        return True

    # tuple[T, ...]
    if len(args) == 2 and args[1] is Ellipsis:
        elem_tp = args[0]
        return all(m.matches(v, elem_tp) for v in value)

    # tuple[T1, T2, ...]
    if len(value) != len(args):
        return False

    # zip(strict=True) would be fine in 3.10+, but keep it simple and explicit
    return all(m.matches(v, elem_tp) for v, elem_tp in zip(value, args, strict=True))



import inspect
import typing
from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from logging import getLogger
from typing import Any, TypeVar

from hassette.exceptions import HassetteNotInitializedError

if typing.TYPE_CHECKING:
    from hassette import Hassette, HassetteConfig, TaskBucket

LOGGER = getLogger(__name__)

CURRENT_BUCKET: ContextVar["TaskBucket | None"] = ContextVar("CURRENT_BUCKET", default=None)
HASSETTE_INSTANCE: ContextVar["Hassette"] = ContextVar("HASSETTE_INSTANCE")
HASSETTE_SET_LOCATION: ContextVar[str | None] = ContextVar("HASSETTE_SET_LOCATION", default=None)
HASSETTE_CONFIG: ContextVar["HassetteConfig"] = ContextVar("HASSETTE_CONFIG")

T = TypeVar("T")

## Setters ##


def set_global_hassette(hassette: "Hassette") -> None:
    """Set the global Hassette instance."""
    curr_inst = HASSETTE_INSTANCE.get(None)
    if curr_inst is hassette:
        return  # already set to the same instance

    if curr_inst is not None:
        extra_msg = f"Set at {HASSETTE_SET_LOCATION.get()}" if HASSETTE_SET_LOCATION.get() else ""
        raise RuntimeError(f"Hassette instance is already set.{extra_msg}")

    try:
        # Capture where this was first set
        frame = inspect.currentframe()
        caller = frame.f_back if frame is not None else None
        if caller is not None:
            info = inspect.getframeinfo(caller)
            where = f"{info.filename}:{info.lineno} in {info.function}"
        else:
            where = "<unknown location>"
    except Exception as e:
        LOGGER.warning("Failed to capture set location for Hassette instance: %s", e)
        where = "<unknown location>"

    HASSETTE_SET_LOCATION.set(where)
    HASSETTE_INSTANCE.set(hassette)


def set_global_hassette_config(config: "HassetteConfig") -> None:
    """Set the global HassetteConfig instance. This can be overriden using the `use` context manager."""
    if HASSETTE_CONFIG.get(None) is not None:
        raise RuntimeError("HassetteConfig is already set in context.")
    HASSETTE_CONFIG.set(config)


## Getters ##


def get_hassette() -> "Hassette":
    """Get the current Hassette instance from context."""
    try:
        inst = HASSETTE_INSTANCE.get()
        return inst
    except LookupError as e:
        raise HassetteNotInitializedError("No Hassette instance found in context.") from e


def get_hassette_config() -> "HassetteConfig":
    """Get the current Hassette configuration from context."""
    try:
        config = HASSETTE_CONFIG.get()
        return config
    except LookupError as e:
        LOGGER.debug("HassetteConfig not found in context, attempting to get from Hassette instance.")
        c = get_hassette().config
        if c is None:
            raise HassetteNotInitializedError("No HassetteConfig found in context or Hassette instance.") from e
        return c


## Context Managers ##


@contextmanager
def use(var: ContextVar[T], value: T) -> Generator[None, Any, Any]:
    """Temporarily set a ContextVar to `value` within a block."""
    token = var.set(value)
    try:
        yield
    finally:
        var.reset(token)


@contextmanager
def use_hassette_config(config: "HassetteConfig") -> Generator[None, Any, Any]:
    """Temporarily set the global HassetteConfig within a block."""
    token = HASSETTE_CONFIG.set(config)
    try:
        yield
    finally:
        HASSETTE_CONFIG.reset(token)


@contextmanager
def use_task_bucket(bucket: "TaskBucket") -> Generator[None, Any, Any]:
    """Temporarily set the current TaskBucket within a block."""
    token = CURRENT_BUCKET.set(bucket)
    try:
        yield
    finally:
        CURRENT_BUCKET.reset(token)



"""API functionality for interacting with Home Assistant.

This module provides clean access to the API classes for making HTTP requests,
managing WebSocket connections, and handling entity states.
"""

from .api import Api
from .sync import ApiSyncFacade

__all__ = ["Api", "ApiSyncFacade"]



from typing import Literal

from pydantic import Field

from .base import AttributesBase, DateTimeBaseState


class SceneAttributes(AttributesBase):
    id: str | None = Field(default=None)


class SceneState(DateTimeBaseState):
    """Representation of a Home Assistant scene state.

    See: https://www.home-assistant.io/integrations/scene/
    """

    domain: Literal["scene"]

    attributes: SceneAttributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class ImageProcessingAttributes(AttributesBase):
    faces: list | None = Field(default=None)
    total_faces: int | float | None = Field(default=None)


class ImageProcessingState(StringBaseState):
    """Representation of a Home Assistant image_processing state.

    See: https://www.home-assistant.io/integrations/image_processing/
    """

    domain: Literal["image_processing"]

    attributes: ImageProcessingAttributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class HumidifierAttributes(AttributesBase):
    min_humidity: float | None = Field(default=None)
    max_humidity: float | None = Field(default=None)
    available_modes: list[str] | None = Field(default=None)
    current_humidity: float | None = Field(default=None)
    humidity: float | None = Field(default=None)
    mode: str | None = Field(default=None)
    action: str | None = Field(default=None)


class HumidifierState(StringBaseState):
    """Representation of a Home Assistant humidifier state.

    See: https://www.home-assistant.io/integrations/humidifier/
    """

    domain: Literal["humidifier"]

    attributes: HumidifierAttributes



from collections import defaultdict, deque
from collections.abc import Iterable
from dataclasses import astuple, dataclass
from typing import Any

from aiohttp import web
from whenever import PlainDateTime

from hassette.utils.request_utils import orjson_dump

# Key = tuple[str, str, str]  # (METHOD, PATH, QUERYSTRING)


@dataclass(eq=True, frozen=True)
class Key:
    method: str
    path: str
    query: str


@dataclass
class Expected:
    status: int
    json: Any


class SimpleTestServer:
    """
    Minimal HTTP double for Home Assistant.

    Usage:
      mock.expect("GET", "/api/states/light.kitchen", "", json={...})
      app.router.add_route("*", "/{tail:.*}", mock.handle_request)
    """

    def __init__(self) -> None:
        self._expectations: dict[Key, deque[Expected]] = defaultdict(deque)
        self._unexpected: list[Key] = []

    # ----- registering expectations -----

    def expect(
        self,
        method: str,
        path: str,
        query: str = "",
        *,
        json: Any = None,
        status: int = 200,
        repeat: int = 1,
    ) -> None:
        key = Key(method.upper(), path, query or "")
        for _ in range(repeat):
            self._expectations[key].append(Expected(status=status, json=json))

    # Nice helper for history endpoints (keeps query ordering stable)
    @staticmethod
    def make_history_path(
        entity_ids: Iterable[str],
        start: PlainDateTime,
        end: PlainDateTime,
        *,
        minimal: bool = False,
    ):
        ids = ",".join(entity_ids)
        path = f"/api/history/period/{start.format_iso()}"
        qs = f"filter_entity_id={ids}&end_time={end.format_iso()}"
        if minimal:
            qs += "&minimal_response=true"
        # Caller still needs to provide METHOD, so this returns (PATH, QUERY)
        return path, qs  # (path, query)

    # ----- request handler -----

    async def handle_request(self, request: web.Request) -> web.StreamResponse:
        key = Key(request.method, request.path, request.query_string or "")
        bucket = self._expectations.get(key)

        if not bucket:
            # record so teardown can fail loudly with details
            self._unexpected.append(key)
            return web.Response(status=599, text=f"Unexpected request: {key}")

        exp = bucket.popleft()
        if exp.json is None:
            return web.Response(status=exp.status)
        return web.json_response(exp.json, status=exp.status, dumps=orjson_dump)

    # ----- teardown assertions -----

    def _leftovers(self) -> list[tuple[Key, int]]:
        return [(k, len(v)) for k, v in self._expectations.items() if v]

    def assert_clean(self) -> None:
        leftovers = self._leftovers()

        errors = []
        if self._unexpected:
            errors.append(f"Unexpected requests: {self._unexpected}")

        if leftovers:
            errors.append(f"Expected requests not seen: {leftovers}")

        assert not errors, f"MockHaApi assertions failed: {errors}"

    def dump_all(self):
        expectations = {astuple(k): [astuple(e) for e in v] for k, v in self._expectations.items()}
        expectations = {str(k): v for k, v in expectations.items() if v}

        # expectations = {str(k): v for k, v in self._expectations.items() if v}
        unexpected = [str(k) for k in self._unexpected]

        return {"expectations": expectations, "unexpected": unexpected}



"""Dependencies are special annotated types that extract data from events.

These are designed to be used in event handlers to automatically extract commonly used
data from events without boilerplate code.

For example, instead of writing:

```python
async def handle_state_change(event: RawStateChangeEvent):
    new_state = event.payload.data.new_state
    # do something with new_state
```

You can use the `NewState` dependency:
```python
from hassette import D, states

async def handle_state_change(new_state: D.StateNew[states.ButtonState]):
    # do something with new_state
```

Hassette will automatically extract the value from the incoming event, cast it to the correct type,
and pass it to your handler.

If you need to write your own dependencies, you can easily do so by annotating
your parameter(s) with `Annotated` and either using an existing accessor from
[accessors][hassette.event_handling.accessors] or writing your own accessor function.

Examples:
    Extracting the new state object from a RawStateChangeEvent
    ```python
    from hassette import D, states

    async def handle_state_change(new_state: D.StateNew[states.ButtonState]):
        # new_state is automatically extracted and typed as states.ButtonState
        print(new_state.state)
    ```

    Extracting the entity_id from any HassEvent
    ```python
    from hassette import D

    async def handle_event(entity_id: D.EntityId):
        # entity_id is automatically extracted
        print(entity_id)
    ```

    Writing your own dependency
    ```python
    from pathlib import Path

    from typing import Annotated
    from hassette import A

    async def handle_event(
        file_path: Annotated[Path, A.get_path("payload.data.changed_file_path")],
    ):
        # do something with file_path
    ```

"""

import typing
from collections.abc import Callable
from dataclasses import dataclass
from typing import Annotated, Any, Generic, TypeAlias, TypeVar

from hassette.const.misc import MISSING_VALUE, FalseySentinel
from hassette.events import Event, HassContext
from hassette.events.hass.hass import TypedStateChangeEvent as ActualTypedStateChangeEvent
from hassette.exceptions import DependencyResolutionError
from hassette.types import StateT

from . import accessors as A

if typing.TYPE_CHECKING:
    from hassette import RawStateChangeEvent  # noqa: F401

T = TypeVar("T", bound=Event[Any])
R = TypeVar("R")

T_Any = TypeVar("T_Any", bound=Any)


@dataclass(slots=True, frozen=True)
class AnnotationDetails(Generic[T]):
    """Details about an annotation used for dependency injection."""

    extractor: Callable[[T], Any]
    """Function to extract the dependency from the event."""

    converter: Callable[[Any, Any], Any] | None = None
    """Optional converter function to convert the extracted value to the desired type."""

    def __post_init__(self):
        if not callable(self.extractor):
            raise TypeError("extractor must be a callable")

        if self.converter is not None and not callable(self.converter):
            raise TypeError("converter must be a callable if provided")


def ensure_present(accessor: Callable[[T], R]) -> Callable[[T], R]:
    """Wrap an accessor to raise if it returns None or MISSING_VALUE.

    Args:
        accessor: The accessor function to wrap

    Returns:
        Wrapped accessor that validates the return value
    """

    def wrapper(event: T) -> R:
        result = accessor(event)

        # Check if the result is None or MISSING_VALUE
        if result is None or result is MISSING_VALUE:
            raise DependencyResolutionError(f"Required dependency returned {type(result).__name__}, expected a value")

        return result

    return wrapper


def identity(x: Any) -> Any:
    """Identity function - returns the input as-is.

    Used when a parameter needs the full event object without transformation.
    """
    return x


# ======================================================================================
# Typed State Change Event
# ======================================================================================
# This annotation converts a RawStateChangeEvent into a TypedStateChangeEvent
# with typed state objects using the StateRegistry.

# Extractor: identity (full event)
# Returns: TypedStateChangeEvent with typed state
TypedStateChangeEvent: TypeAlias = Annotated[
    ActualTypedStateChangeEvent[StateT], AnnotationDetails["RawStateChangeEvent"](identity)
]
"""Convert a RawStateChangeEvent into a TypedStateChangeEvent with typed state objects.

Example:
```python
async def handler(event: D.TypedStateChangeEvent[states.LightState]):
    brightness = event.payload.data.new_state.attributes.brightness
```
"""


# ======================================================================================
# State Object Extractors
# ======================================================================================
# These annotations extract full state objects (dicts) from events and convert them
# to typed Pydantic models using the StateRegistry.

# Extractor: get_state_object_new() -> HassStateDict
# Returns: Typed state model, raises if None/MISSING_VALUE
StateNew: TypeAlias = Annotated[
    StateT,
    AnnotationDetails["RawStateChangeEvent"](ensure_present(A.get_state_object_new)),
]
"""Extract the new state object from a StateChangeEvent.

Example:
```python
async def handler(new_state: D.StateNew[states.LightState]):
    brightness = new_state.attributes.brightness
```
"""

# Extractor: get_state_object_new() -> HassStateDict | None
# Returns: Typed state model or None
MaybeStateNew: TypeAlias = Annotated[
    StateT | None,
    AnnotationDetails["RawStateChangeEvent"](A.get_state_object_new),
]
"""Extract the new state object from a StateChangeEvent, allowing for None.

Example:
```python
async def handler(new_state: D.MaybeStateNew[states.LightState]):
    if new_state:
        brightness = new_state.attributes.brightness
```
"""

# Extractor: get_state_object_old() -> HassStateDict
# Returns: Typed state model, raises if None/MISSING_VALUE
StateOld: TypeAlias = Annotated[
    StateT,
    AnnotationDetails["RawStateChangeEvent"](ensure_present(A.get_state_object_old)),
]
"""Extract the old state object from a StateChangeEvent.

Example:
```python
async def handler(old_state: D.StateOld[states.LightState]):
    if old_state:
        previous_brightness = old_state.attributes.brightness
```
"""

# Extractor: get_state_object_old() -> HassStateDict | None
# Returns: Typed state model or None
MaybeStateOld: TypeAlias = Annotated[
    StateT | None,
    AnnotationDetails["RawStateChangeEvent"](A.get_state_object_old),
]
"""Extract the old state object from a StateChangeEvent, allowing for None.

Example:
```python
async def handler(old_state: D.MaybeStateOld[states.LightState]):
    if old_state:
        previous_brightness = old_state.attributes.brightness
```
"""


# ======================================================================================
# Identity & Metadata Extractors
# ======================================================================================
# These annotations extract simple identity and metadata fields from events.

# Extractor: get_entity_id() -> str
# Returns: Entity ID string, raises if None/MISSING_VALUE
EntityId: TypeAlias = Annotated[str, AnnotationDetails(ensure_present(A.get_entity_id))]
"""Extract the entity_id from a HassEvent.

Returns the entity ID string (e.g., "light.bedroom").

Example:
```python
async def handler(entity_id: D.EntityId):
    self.logger.info("Entity: %s", entity_id)
```
"""

# Extractor: get_entity_id() -> str | FalseySentinel
# Returns: Entity ID string or MISSING_VALUE
MaybeEntityId: TypeAlias = Annotated[str | FalseySentinel, AnnotationDetails(A.get_entity_id)]
"""Extract the entity_id from a HassEvent, returning MISSING_VALUE sentinel if not present."""

# Extractor: get_domain() -> str
# Returns: Domain string, raises if None/MISSING_VALUE
Domain: TypeAlias = Annotated[str, AnnotationDetails(ensure_present(A.get_domain))]
"""Extract the domain from a HassEvent.

Returns the domain string (e.g., "light", "sensor") from the event payload or entity_id.

Example:
```python

async def handler(domain: D.Domain):
    if domain == "light":
        self.logger.info("Light entity event")
```
"""
# Extractor: get_domain() -> str | FalseySentinel
# Returns: Domain string or MISSING_VALUE
MaybeDomain: TypeAlias = Annotated[str | FalseySentinel, AnnotationDetails(A.get_domain)]
"""Extract the domain from a HassEvent, returning MISSING_VALUE sentinel if not present."""

# Extractor: get_context() -> HassContext
# Returns: HassContext object
EventContext: TypeAlias = Annotated[HassContext, AnnotationDetails[Event](A.get_context)]
"""Extract the context object from a HassEvent.

Returns the Home Assistant context object containing metadata about the event
origin (user_id, parent_id, etc.).

Example:
```python
async def handler(context: D.EventContext):
    if context.user_id:
        self.logger.info("Triggered by user: %s", context.user_id)
```
"""



from typing import Literal

from pydantic import Field
from whenever import Instant, PlainDateTime, ZonedDateTime

from .base import AttributesBase, StringBaseState


class CalendarAttributes(AttributesBase):
    message: str | None = Field(default=None)
    all_day: bool | None = Field(default=None)
    start_time: Instant | PlainDateTime | ZonedDateTime | None = Field(default=None)
    end_time: Instant | PlainDateTime | ZonedDateTime | None = Field(default=None)
    location: str | None = Field(default=None)
    description: str | None = Field(default=None)
    offset_reached: bool | None = Field(default=None)


class CalendarState(StringBaseState):
    """Representation of a Home Assistant calendar state.

    See: https://www.home-assistant.io/integrations/calendar/
    """

    domain: Literal["calendar"]

    attributes: CalendarAttributes



from enum import StrEnum, auto


class Topic(StrEnum):
    # hassette events

    HASSETTE_EVENT_SERVICE_STATUS = "hassette.event.service_status"
    """Service status updates"""

    HASSETTE_EVENT_WEBSOCKET_CONNECTED = "hassette.event.websocket_connected"
    """WebSocket connection established"""

    HASSETTE_EVENT_WEBSOCKET_DISCONNECTED = "hassette.event.websocket_disconnected"
    """WebSocket connection lost"""

    HASSETTE_EVENT_FILE_WATCHER = "hassette.event.file_watcher"
    """File watcher events"""

    HASSETTE_EVENT_APP_LOAD_COMPLETED = "hassette.event.app_load_completed"
    """Application load completion events"""

    # Home Assistant events

    HASS_EVENT_STATE_CHANGED = "hass.event.state_changed"
    """State change events"""

    HASS_EVENT_CALL_SERVICE = "hass.event.call_service"
    """Service call events"""

    HASS_EVENT_COMPONENT_LOADED = "hass.event.component_loaded"
    """Component loaded events"""

    HASS_EVENT_SERVICE_REGISTERED = "hass.event.service_registered"
    """Service registered events"""

    HASS_EVENT_SERVICE_REMOVED = "hass.event.service_removed"
    """Service removed events"""

    HASS_EVENT_LOGBOOK_ENTRY = "hass.event.logbook_entry"
    """Logbook entry events"""

    HASS_EVENT_USER_ADDED = "hass.event.user_added"
    """User added events"""

    HASS_EVENT_USER_REMOVED = "hass.event.user_removed"
    """User removed events"""

    HASS_EVENT_AUTOMATION_TRIGGERED = "hass.event.automation_triggered"
    """Automation triggered events"""

    HASS_EVENT_SCRIPT_STARTED = "hass.event.script_started"
    """Script started events"""


class ResourceStatus(StrEnum):
    """Enumeration for resource status."""

    NOT_STARTED = auto()
    """The resource has not been started yet."""

    STARTING = auto()
    """The resource is in the process of starting."""

    RUNNING = auto()
    """The resource is currently running."""

    STOPPED = auto()
    """The resource has been stopped without errors."""

    FAILED = auto()
    """The resource has failed with a recoverable error."""

    CRASHED = auto()
    """The resource has crashed unexpectedly and cannot recover."""


class ResourceRole(StrEnum):
    """Enumeration for resource roles."""

    CORE = "Core"
    """Only used by Hassette directly, as it does not inherit from Resource."""

    BASE = "Base"
    """The base role for all resources."""

    SERVICE = "Service"
    """A service resource."""

    RESOURCE = "Resource"
    """A generic resource."""

    APP = "App"
    """An application resource."""

    UNKNOWN = "Unknown"
    """An unknown or unclassified resource."""



import asyncio
import typing
from collections.abc import Coroutine
from typing import Any, Protocol, TypeVar

from hassette.types.enums import ResourceStatus

if typing.TYPE_CHECKING:
    import logging

    from hassette.events import HassetteServiceEvent

T = TypeVar("T")
CoroLikeT = Coroutine[Any, Any, T]


class _TaskBucketP(Protocol):
    def spawn(self, coro: CoroLikeT, *, name: str | None = None) -> asyncio.Task: ...
    async def cancel_all(self) -> None: ...


class _HassetteP(Protocol):
    async def send_event(self, topic: str, payload: Any) -> None: ...


# shim for typing only - LifecycleMixin needs these attributes to be present
# but we don't want to enforce inheritance from Resource or HassetteBase at runtime
if typing.TYPE_CHECKING:

    class _LifecycleHostStubs(Protocol):
        logger: logging.Logger
        hassette: _HassetteP
        role: Any
        class_name: str
        unique_name: str
        task_bucket: _TaskBucketP

        def _create_service_status_event(
            self, status: ResourceStatus, exception: Exception | None = None
        ) -> "HassetteServiceEvent": ...

        async def initialize(self, *args, **kwargs) -> None: ...
else:

    class _LifecycleHostStubs:  # runtime stub (empty)
        pass


class LifecycleMixin(_LifecycleHostStubs):
    ready_event: asyncio.Event
    """Event to signal readiness of the instance."""

    shutdown_event: asyncio.Event
    """Event to signal shutdown of the instance."""

    _ready_reason: str | None
    """Optional reason for readiness or lack thereof."""

    _init_task: asyncio.Task | None = None
    """Initialization task for the instance."""

    _previous_status: ResourceStatus = ResourceStatus.NOT_STARTED
    """Previous status of the instance."""

    _status: ResourceStatus = ResourceStatus.NOT_STARTED
    """Current status of the instance."""

    def __init__(self) -> None:
        self.ready_event = asyncio.Event()
        self.shutdown_event = asyncio.Event()
        self._ready_reason = None
        self._previous_status = ResourceStatus.NOT_STARTED
        self._status = ResourceStatus.NOT_STARTED
        self._init_task: asyncio.Task | None = None

    # --------- props
    @property
    def status(self) -> ResourceStatus:
        return self._status

    @status.setter
    def status(self, value: ResourceStatus) -> None:
        self._previous_status = self._status
        self._status = value

    @property
    def task(self) -> asyncio.Task | None:
        return self._init_task

    # --------- lifecycle ops
    def start(self) -> None:
        """Start the instance by spawning its initialize method in a task."""
        # create a new event each time we start
        self.shutdown_event = asyncio.Event()

        if self._init_task and not self._init_task.done():
            self.logger.debug("%s already started or running", self.unique_name, stacklevel=2)
            return

        self.logger.debug("%s starting", self.unique_name)
        self._init_task = self.task_bucket.spawn(self.initialize(), name="resource:resource_initialize")

    def cancel(self) -> None:
        """Cancel the main task of the instance, if it is running."""
        if self._init_task and not self._init_task.done():
            self._init_task.cancel()
            self.logger.debug("%s cancelled task", self.unique_name)
            return

        self.logger.debug("%s no running task to cancel", self.unique_name)

    # --------- readiness
    def mark_ready(self, reason: str | None = None) -> None:
        """Mark the instance as ready.

        Args:
            reason: Optional reason for readiness.
        """
        if self.ready_event.is_set():
            self.logger.debug("%s already ready, skipping reason %s", self.unique_name, reason)
            return
        self.logger.debug("ready: %s", reason or "no reason provided")
        self._ready_reason = reason
        self.ready_event.set()

    def mark_not_ready(self, reason: str | None = None) -> None:
        """Mark the instance as not ready.

        Args:
            reason: Optional reason for lack of readiness.
        """
        if not self.ready_event.is_set():
            self.logger.debug("%s already not ready, skipping reason %s", self.unique_name, reason)

        self._ready_reason = reason
        self.ready_event.clear()

    def request_shutdown(self, reason: str | None = None) -> None:
        """Set the sticky shutdown flag. Idempotent."""
        if not self.shutdown_event.is_set():
            self.logger.debug("%s shutdown requested: %s", self.unique_name, reason or "")
            self.shutdown_event.set()
        # clear readiness early so callers back off
        self.mark_not_ready(reason or "shutdown requested")

    def is_ready(self) -> bool:
        """Check if the instance is ready.

        Returns:
            True if the instance is ready, False otherwise.
        """
        return self.ready_event.is_set()

    async def wait_ready(self, timeout: float | None = None) -> None:
        """Wait until the instance is marked as ready.

        Args:
            timeout: Optional timeout in seconds to wait for readiness. If None, wait indefinitely.

        Raises:
            TimeoutError: If the timeout is reached before the instance is ready.
        """
        if timeout is None:
            await self.ready_event.wait()
        else:
            await asyncio.wait_for(self.ready_event.wait(), timeout)

    # --------- transitions
    async def handle_stop(self) -> None:
        if self.status == ResourceStatus.STOPPED:
            self.logger.debug("%s already stopped", self.unique_name, stacklevel=2)
            return

        self.logger.debug("%s stopping", self.unique_name, stacklevel=2)
        self.status = ResourceStatus.STOPPED
        event = self._create_service_status_event(ResourceStatus.STOPPED)
        await self.hassette.send_event(event.topic, event)
        self.mark_not_ready("Stopped")

    async def handle_failed(self, exception: Exception | BaseException) -> None:
        if self.status == ResourceStatus.FAILED:
            self.logger.debug("%s already in failed state", self.unique_name, stacklevel=2)
            return

        self.logger.error("%s failed: %s - %s", self.unique_name, type(exception).__name__, str(exception))
        self.status = ResourceStatus.FAILED
        event = self._create_service_status_event(ResourceStatus.FAILED, exception)
        await self.hassette.send_event(event.topic, event)
        self.mark_not_ready("Failed")

    async def handle_running(self) -> None:
        if self.status == ResourceStatus.RUNNING:
            self.logger.debug("%s already running", self.unique_name, stacklevel=2)
            return

        self.logger.debug("%s running", self.unique_name, stacklevel=2)
        self.status = ResourceStatus.RUNNING
        event = self._create_service_status_event(ResourceStatus.RUNNING)
        await self.hassette.send_event(event.topic, event)

    async def handle_starting(self) -> None:
        if self.status == ResourceStatus.STARTING:
            self.logger.debug("%s already starting", self.unique_name, stacklevel=2)
            return
        self.logger.debug("%s starting", self.unique_name, stacklevel=2)
        self.status = ResourceStatus.STARTING
        event = self._create_service_status_event(ResourceStatus.STARTING)
        await self.hassette.send_event(event.topic, event)

    async def handle_crash(self, exception: Exception) -> None:
        if self.status == ResourceStatus.CRASHED:
            self.logger.debug("%s already in crashed state", self.unique_name, stacklevel=2)
            return

        self.logger.error("%s crashed: %s - %s", self.unique_name, type(exception).__name__, str(exception))
        self.status = ResourceStatus.CRASHED
        event = self._create_service_status_event(ResourceStatus.CRASHED, exception)
        await self.hassette.send_event(event.topic, event)
        self.mark_not_ready("Crashed")

    def _create_service_status_event(self, status: ResourceStatus, exception: Exception | BaseException | None = None):
        from hassette.events import HassetteServiceEvent

        return HassetteServiceEvent.from_data(
            resource_name=self.class_name,
            role=self.role,
            status=status,
            previous_status=self._previous_status,
            exception=exception,
        )



"""Type stub file for States resource.

This stub provides static type hints for all known domain properties on the States
class, enabling IDE autocomplete and type checking. At runtime, these properties are
handled by __getattr__ in states.py, which queries the state registry dynamically.

For custom domains not listed here, users should use:
```python
    app.states[CustomStateClass]
```

As this will ensure proper type annotations of the return type.
"""

import typing
from collections.abc import Iterator
from typing import Generic, NamedTuple

from frozendict import frozendict

from hassette.conversion import StateKey
from hassette.models import states
from hassette.resources.base import Resource
from hassette.types import StateT

if typing.TYPE_CHECKING:
    from hassette import Hassette
    from hassette.core.state_proxy import StateProxy
    from hassette.events import HassStateDict

class CacheValue(Generic[StateT], NamedTuple):
    context_id: str | None
    frozen_state: frozendict
    model: StateT

class DomainStates(typing.Generic[StateT]):
    _state_proxy: StateProxy
    _model: type[StateT]
    _domain: str
    _cache: dict[str, CacheValue[StateT]]

    def _validate_or_return_from_cache(self, entity_id: str, state: HassStateDict) -> StateT: ...
    def __init__(self, state_proxy: StateProxy, model: type[StateT]) -> None: ...
    def __iter__(self) -> typing.Generator[tuple[str, StateT], typing.Any, None]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, entity_id: str) -> StateT: ...
    def __contains__(self, entity_id: str) -> bool: ...
    def __repr__(self) -> str: ...
    def __bool__(self) -> bool: ...
    def get(self, entity_id: str) -> StateT | None: ...
    def items(self) -> Iterator[tuple[str, StateT]]: ...
    def keys(self) -> list[str]: ...
    def iterkeys(self) -> typing.Iterator[str]: ...
    def values(self) -> list[StateT]: ...
    def itervalues(self) -> Iterator[StateT]: ...
    def to_dict(self) -> dict[str, StateT]: ...

class StateManager(Resource):
    @property
    def _state_proxy(self) -> StateProxy: ...
    @classmethod
    def create(cls, hassette: Hassette, parent: Resource) -> StateManager: ...

    # Known domain properties with full typing
    @property
    def ai_task(self) -> DomainStates[states.AiTaskState]: ...
    @property
    def air_quality(self) -> DomainStates[states.AirQualityState]: ...
    @property
    def alarm_control_panel(self) -> DomainStates[states.AlarmControlPanelState]: ...
    @property
    def assist_satellite(self) -> DomainStates[states.AssistSatelliteState]: ...
    @property
    def automation(self) -> DomainStates[states.AutomationState]: ...
    @property
    def binary_sensor(self) -> DomainStates[states.BinarySensorState]: ...
    @property
    def button(self) -> DomainStates[states.ButtonState]: ...
    @property
    def calendar(self) -> DomainStates[states.CalendarState]: ...
    @property
    def camera(self) -> DomainStates[states.CameraState]: ...
    @property
    def climate(self) -> DomainStates[states.ClimateState]: ...
    @property
    def conversation(self) -> DomainStates[states.ConversationState]: ...
    @property
    def cover(self) -> DomainStates[states.CoverState]: ...
    @property
    def date(self) -> DomainStates[states.DateState]: ...
    @property
    def datetime(self) -> DomainStates[states.DateTimeState]: ...
    @property
    def device_tracker(self) -> DomainStates[states.DeviceTrackerState]: ...
    @property
    def event(self) -> DomainStates[states.EventState]: ...
    @property
    def fan(self) -> DomainStates[states.FanState]: ...
    @property
    def humidifier(self) -> DomainStates[states.HumidifierState]: ...
    @property
    def image_processing(self) -> DomainStates[states.ImageProcessingState]: ...
    @property
    def input_boolean(self) -> DomainStates[states.InputBooleanState]: ...
    @property
    def input_button(self) -> DomainStates[states.InputButtonState]: ...
    @property
    def input_datetime(self) -> DomainStates[states.InputDatetimeState]: ...
    @property
    def input_number(self) -> DomainStates[states.InputNumberState]: ...
    @property
    def input_select(self) -> DomainStates[states.InputSelectState]: ...
    @property
    def input_text(self) -> DomainStates[states.InputTextState]: ...
    @property
    def light(self) -> DomainStates[states.LightState]: ...
    @property
    def lock(self) -> DomainStates[states.LockState]: ...
    @property
    def media_player(self) -> DomainStates[states.MediaPlayerState]: ...
    @property
    def notify(self) -> DomainStates[states.NotifyState]: ...
    @property
    def number(self) -> DomainStates[states.NumberState]: ...
    @property
    def person(self) -> DomainStates[states.PersonState]: ...
    @property
    def remote(self) -> DomainStates[states.RemoteState]: ...
    @property
    def scene(self) -> DomainStates[states.SceneState]: ...
    @property
    def script(self) -> DomainStates[states.ScriptState]: ...
    @property
    def select(self) -> DomainStates[states.SelectState]: ...
    @property
    def sensor(self) -> DomainStates[states.SensorState]: ...
    @property
    def siren(self) -> DomainStates[states.SirenState]: ...
    @property
    def stt(self) -> DomainStates[states.SttState]: ...
    @property
    def sun(self) -> DomainStates[states.SunState]: ...
    @property
    def switch(self) -> DomainStates[states.SwitchState]: ...
    @property
    def text(self) -> DomainStates[states.TextState]: ...
    @property
    def time(self) -> DomainStates[states.TimeState]: ...
    @property
    def timer(self) -> DomainStates[states.TimerState]: ...
    @property
    def todo(self) -> DomainStates[states.TodoState]: ...
    @property
    def tts(self) -> DomainStates[states.TtsState]: ...
    @property
    def update(self) -> DomainStates[states.UpdateState]: ...
    @property
    def vacuum(self) -> DomainStates[states.VacuumState]: ...
    @property
    def valve(self) -> DomainStates[states.ValveState]: ...
    @property
    def water_heater(self) -> DomainStates[states.WaterHeaterState]: ...
    @property
    def weather(self) -> DomainStates[states.WeatherState]: ...
    @property
    def zone(self) -> DomainStates[states.ZoneState]: ...
    def __getitem__(self, model: type[StateT]) -> DomainStates[StateT]: ...
    def __contains__(self, model: type[StateT]) -> bool: ...
    def __iter__(self) -> Iterator[tuple[StateKey, DomainStates[states.BaseState]]]: ...
    def items(self) -> Iterator[tuple[StateKey, DomainStates[states.BaseState]]]: ...
    def values(self) -> Iterator[DomainStates[states.BaseState]]: ...
    def keys(self) -> Iterator[StateKey]: ...



from typing import Any, Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class UpdateAttributes(AttributesBase):
    auto_update: bool | None = Field(default=None)
    display_precision: int | float | None = Field(default=None)
    installed_version: str | None = Field(default=None)
    in_progress: bool | None = Field(default=None)
    latest_version: str | None = Field(default=None)
    release_summary: Any | None = Field(default=None)
    release_url: str | None = Field(default=None)
    skipped_version: Any | None = Field(default=None)
    title: str | None = Field(default=None)
    update_percentage: Any | None = Field(default=None)
    entity_picture: str | None = Field(default=None)


class UpdateState(StringBaseState):
    """Representation of a Home Assistant update state.

    See: https://www.home-assistant.io/integrations/update/
    """

    domain: Literal["update"]

    attributes: UpdateAttributes



import itertools
import typing
from dataclasses import dataclass, field
from typing import Generic, Literal, TypeVar

from whenever import ZonedDateTime

from hassette.types import PayloadT

if typing.TYPE_CHECKING:
    from hassette.types import Topic

DataT = TypeVar("DataT", covariant=True)
"""Represents the data type within an event payload."""


HASSETTE_EVENT_ID_SEQ = itertools.count(1)


@dataclass(frozen=True, slots=True)
class EventPayload(Generic[DataT]):
    """Base payload with typed data."""

    event_type: str
    """Type of the event."""

    data: DataT
    """The actual event data."""


@dataclass(frozen=True, slots=True)
class HassContext:
    """Structure for the context of a Home Assistant event."""

    id: str
    parent_id: str | None
    user_id: str | None


@dataclass(frozen=True, slots=True, repr=False)
class HassPayload(EventPayload[DataT]):
    """Home Assistant event payload with additional metadata."""

    event_type: str
    """Type of the event, e.g., 'state_changed', 'call_service', etc."""

    data: DataT
    """The actual event data from Home Assistant."""

    origin: Literal["LOCAL", "REMOTE"]
    """Origin of the event, either 'LOCAL' or 'REMOTE'."""

    time_fired: ZonedDateTime
    """The time the event was fired."""

    context: HassContext
    """The context of the event."""

    @property
    def entity_id(self) -> str | None:
        """Return the entity_id if present in the data."""
        return getattr(self.data, "entity_id", None)

    @property
    def domain(self) -> str | None:
        """Return the domain if present in the data."""
        if hasattr(self.data, "domain"):
            return getattr(self.data, "domain", None)

        entity_id = self.entity_id
        if entity_id:
            return entity_id.split(".")[0]
        return None

    @property
    def service(self) -> str | None:
        """Return the service if present in the data."""
        return getattr(self.data, "service", None)

    @property
    def event_id(self) -> str:
        """The unique identifier for the event."""
        return self.context.id

    def __repr__(self) -> str:
        if self.entity_id:
            return f"HassPayload(event_type={self.event_type}, entity_id={self.entity_id}, event_id={self.event_id})"

        return f"HassPayload(event_type={self.event_type}, event_id={self.event_id})"


@dataclass(frozen=True, slots=True, repr=False)
class HassettePayload(EventPayload[DataT]):
    """Hassette event payload with additional metadata."""

    event_type: str
    """Type of the event, e.g., 'state_changed', 'call_service', etc."""

    data: DataT
    """The actual event data from Home Assistant."""

    event_id: int = field(default_factory=lambda: next(HASSETTE_EVENT_ID_SEQ))
    """The unique identifier for the event."""

    def __repr__(self) -> str:
        return f"HassettePayload(event_type={self.event_type}, event_id={self.event_id})"


@dataclass(frozen=True, slots=True, repr=False)
class Event(Generic[PayloadT]):
    """Base event with strongly typed payload."""

    topic: "Topic | str"
    """Topic of the event."""

    payload: PayloadT
    """The event payload."""

    def __repr__(self) -> str:
        return f"Event({self.payload})"



import asyncio
import contextlib
import logging
import threading
import typing
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, cast
from unittest.mock import AsyncMock, Mock, PropertyMock, patch

from aiohttp import web
from anyio import create_memory_object_stream
from yarl import URL

from hassette import HassetteConfig, context
from hassette.api import Api
from hassette.bus import Bus
from hassette.conversion import STATE_REGISTRY, TYPE_REGISTRY, StateRegistry, TypeRegistry
from hassette.core.api_resource import ApiResource
from hassette.core.app_handler import AppHandler
from hassette.core.bus_service import BusService
from hassette.core.file_watcher import FileWatcherService
from hassette.core.scheduler_service import SchedulerService
from hassette.core.state_proxy import StateProxy
from hassette.core.websocket_service import WebsocketService
from hassette.events import Event
from hassette.resources.base import Resource
from hassette.scheduler import Scheduler
from hassette.state_manager import StateManager
from hassette.task_bucket import TaskBucket, make_task_factory
from hassette.test_utils.test_server import SimpleTestServer
from hassette.types.enums import ResourceStatus
from hassette.utils.service_utils import wait_for_ready
from hassette.utils.url_utils import build_rest_url, build_ws_url

if typing.TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

    from hassette import Hassette


async def wait_for(
    predicate: Callable[[], bool], *, timeout: float = 3.0, interval: float = 0.02, desc: str = "condition"
) -> None:
    loop = asyncio.get_running_loop()
    deadline = loop.time() + timeout
    while True:
        if predicate():
            return
        if loop.time() >= deadline:
            raise TimeoutError(f"Timed out waiting for {desc}")
        await asyncio.sleep(interval)


async def start_resource(res: Resource, *, desc: str) -> asyncio.Task[Any] | None:
    res.start()
    task: asyncio.Task[Any] | None = None
    task = res.task
    await wait_for(lambda: getattr(res, "status", None) == ResourceStatus.RUNNING, desc=f"{desc} RUNNING")
    return task


async def shutdown_resource(res: Resource) -> None:
    with contextlib.suppress(Exception):
        await res.shutdown()


class _HassetteMock(Resource):
    task_bucket: TaskBucket

    def __init__(self, *, config: HassetteConfig) -> None:
        self.config = config
        super().__init__(cast("Hassette", self))

        self.ready_event = asyncio.Event()
        self.shutdown_event = asyncio.Event()
        self.children: list[Resource] = []
        self._loop: asyncio.AbstractEventLoop | None = None
        self._loop_thread_id: int | None = None
        self._send_stream: MemoryObjectSendStream[tuple[str, Event[Any]]] | None = None
        self._receive_stream: MemoryObjectReceiveStream[tuple[str, Event[Any]]] | None = None

        self._api_service: ApiResource | None = None
        self.api: Api | None = None
        self._bus_service: BusService | None = None
        self._bus: Bus | None = None
        self._scheduler_service: SchedulerService | None = None
        self._scheduler: Scheduler | None = None
        self._file_watcher: FileWatcherService | None = None
        self._app_handler: AppHandler | None = None
        self._websocket_service: WebsocketService | None = None
        self._state_proxy: StateProxy | None = None
        self._states: StateManager | None = None
        self.state_registry: StateRegistry | None = None
        self.type_registry: TypeRegistry | None = None

    @property
    def ws_url(self) -> str:
        """Construct the WebSocket URL for Home Assistant."""
        return build_ws_url(self.config)

    @property
    def rest_url(self) -> str:
        """Construct the REST API URL for Home Assistant."""
        return build_rest_url(self.config)

    async def send_event(self, topic: str, event: Event[Any]) -> None:
        if not self._send_stream:
            raise RuntimeError("Bus is not enabled on this harness")
        if self._send_stream._closed:
            raise RuntimeError("Event bus send stream is closed")
        await self._send_stream.send((topic, event))

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not self._loop:
            raise RuntimeError("Event loop is not running")
        return self._loop

    async def wait_for_ready(self, resources: list[Resource] | Resource, timeout: int | None = None) -> bool:
        """Block until all dependent resources are ready or shutdown is requested.

        Args:
            resources: The resource(s) to wait for.
            timeout: The timeout for the wait operation.

        Returns:
            True if all resources are ready, False if shutdown is requested.
        """
        timeout = timeout or self.config.startup_timeout_seconds

        return await wait_for_ready(resources, timeout=timeout, shutdown_event=self.shutdown_event)

    def get_app(self, name: str, index: int = 0) -> Any:
        if not self._app_handler:
            raise RuntimeError("App handler is not enabled on this harness")
        return self._app_handler.get(name, index=index)

    @property
    def event_streams_closed(self) -> bool:
        """Check if the event streams are closed."""
        if not self._send_stream or not self._receive_stream:
            return True
        return self._send_stream._closed and self._receive_stream._closed


@dataclass
class HassetteHarness:
    config: HassetteConfig
    use_bus: bool = False
    use_scheduler: bool = False
    use_api_mock: bool = False
    use_api_real: bool = False
    use_file_watcher: bool = False
    use_app_handler: bool = False
    use_websocket: bool = False
    use_state_proxy: bool = False
    use_state_registry: bool = False
    unused_tcp_port: int = 0

    def __post_init__(self) -> None:
        # need this for caplog to work properly
        logging.getLogger("hassette").propagate = True

        if self.use_api_mock and self.use_api_real:
            raise ValueError("Cannot use both API mock and real API in the same harness")

        self.logger = logging.getLogger(f"hassette.test.harness.{type(self).__name__}")
        self.hassette = _HassetteMock(config=self.config)
        self._tasks: list[tuple[str, asyncio.Task[Any]]] = []
        self._exit_stack = contextlib.AsyncExitStack()
        self.api_mock: SimpleTestServer | None = None
        self.api_base_url = URL.build(scheme="http", host="127.0.0.1", port=self.unused_tcp_port, path="/api/")

        context.set_global_hassette(cast("Hassette", self.hassette))

        self.config.set_validated_app_manifests()

    async def __aenter__(self):
        await self.start()

        assert self.hassette._loop is not None, "Event loop is not running"
        assert self.hassette._loop.is_running(), "Event loop is not running"
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.stop()

    async def start(self) -> "HassetteHarness":
        self.hassette._loop = asyncio.get_running_loop()
        self.hassette._loop_thread_id = threading.get_ident()
        self.hassette.task_bucket = TaskBucket.create(cast("Hassette", self.hassette), parent=self.hassette)  # pyright: ignore[reportArgumentType]
        self.hassette._loop.set_task_factory(make_task_factory(self.hassette.task_bucket))  # pyright: ignore[reportArgumentType]

        if self.use_bus:
            await self._start_bus()
        if self.use_scheduler:
            await self._start_scheduler()
        if self.use_file_watcher:
            await self._start_file_watcher()
        if self.use_api_mock:
            await self._start_api_mock()

        # if self.use_api_real:
        #     await self.start_api_real()
        if self.use_app_handler:
            await self._start_app_handler()
            if not self.use_state_proxy:
                await self._start_state_proxy()
        # if self.use_websocket:
        #     await self.start_websocket()

        # Set up API and websocket mocks before state proxy if needed
        if not self.hassette._api_service:
            self.hassette._api_service = Mock()
            self.hassette._api_service.ready_event = asyncio.Event()
            self.hassette._api_service.ready_event.set()

        if not self.hassette._websocket_service:
            self.hassette._websocket_service = Mock()
            self.hassette._websocket_service.ready_event = asyncio.Event()
            self.hassette._websocket_service.ready_event.set()

        if not self.hassette.api:
            self.hassette.api = AsyncMock()
            self.hassette.api.sync = Mock()

        if self.use_state_proxy:
            if not self.use_bus:
                raise RuntimeError("State proxy requires bus")
            await self._start_state_proxy()

        if self.use_state_registry:
            self.hassette.state_registry = STATE_REGISTRY
            self.hassette.type_registry = TYPE_REGISTRY

        self.hassette._states = self.hassette.add_child(StateManager)

        if not self.use_bus:
            self.hassette.send_event = AsyncMock()

        for resource in self.hassette.children:
            resource.start()

        self.hassette.ready_event.set()
        await wait_for_ready(
            [x for x in self.hassette.children], timeout=1, shutdown_event=self.hassette.shutdown_event
        )

        return self

    async def stop(self) -> None:
        self.hassette.shutdown_event.set()

        try:
            for resource in self.hassette.children:
                await shutdown_resource(resource)
                await asyncio.sleep(0.05)
        except Exception:
            self.logger.exception("Error shutting down resources")

        try:
            for _, task in reversed(self._tasks):
                task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await task
        except Exception:
            self.logger.exception("Error cancelling tasks")

        try:
            if self.api_mock is not None:
                self.api_mock.assert_clean()
        except Exception:
            self.logger.exception("Error checking API mock")

        await self._exit_stack.aclose()

        self.hassette._loop = None

    async def _start_bus(self) -> None:
        send_stream, receive_stream = create_memory_object_stream[tuple[str, Event[Any]]](1000)
        self.hassette._send_stream = send_stream
        self.hassette._receive_stream = receive_stream

        self.hassette._bus_service = self.hassette.add_child(BusService, stream=receive_stream.clone())
        self.hassette._bus = self.hassette.add_child(Bus)

        self._exit_stack.push_async_callback(send_stream.aclose)
        self._exit_stack.push_async_callback(receive_stream.aclose)

    async def _start_scheduler(self) -> None:
        self.hassette._scheduler_service = self.hassette.add_child(SchedulerService)
        self.hassette._scheduler = self.hassette.add_child(Scheduler)

    async def _start_file_watcher(self) -> None:
        if not self.hassette.config:
            raise RuntimeError("File watcher requires a config")

        self.hassette._file_watcher = self.hassette.add_child(FileWatcherService)

    async def _start_app_handler(self) -> None:
        if not self.use_bus:
            raise RuntimeError("App handler requires bus")

        self.hassette._app_handler = self.hassette.add_child(AppHandler)
        self.hassette._websocket_service = Mock()
        self.hassette._websocket_service.status = ResourceStatus.RUNNING

        return

    async def _start_api_mock(self) -> None:
        if not self.api_base_url:
            raise RuntimeError("API harness requires api_base_url")
        mock_server = SimpleTestServer()
        app = web.Application()
        app.router.add_route("*", "/{tail:.*}", mock_server.handle_request)

        runner = web.AppRunner(app)
        await runner.setup()
        self._exit_stack.push_async_callback(runner.cleanup)

        site = web.TCPSite(runner, self.api_base_url.host or "127.0.0.1", self.api_base_url.port or 80)
        await site.start()
        self._exit_stack.push_async_callback(site.stop)

        rest_url_patch = patch(
            "hassette.core.api_resource.ApiResource._rest_url",
            new_callable=PropertyMock,
            return_value=self.api_base_url,
        )
        headers_patch = patch(
            "hassette.core.api_resource.ApiResource._headers",
            new_callable=PropertyMock,
            return_value={"Authorization": "Bearer test_token"},
        )
        self._exit_stack.enter_context(rest_url_patch)
        self._exit_stack.enter_context(headers_patch)

        self.hassette._websocket_service = Mock(spec=WebsocketService)
        self.hassette._websocket_service.ready_event = asyncio.Event()
        self.hassette._websocket_service.ready_event.set()

        self.hassette._api_service = self.hassette.add_child(ApiResource)
        self.hassette.api = self.hassette.add_child(Api)

        self.api_mock = mock_server

        return

    async def _start_state_proxy(self) -> None:
        if not self.use_bus:
            raise RuntimeError("State proxy requires bus")

        self.hassette._state_proxy = self.hassette.add_child(StateProxy)
        return



import typing

from aiohttp import web

from hassette.resources.base import Service
from hassette.types.enums import ResourceStatus

if typing.TYPE_CHECKING:
    from hassette import Hassette

_T = typing.TypeVar("_T")


# subclass to prevent the weird UnboundLocalError we get from aiohttp
# i think it's due to pytest but i'm tired of trying to figure it out
# that's why you don't use frame inspections
class MyAppKey(web.AppKey[_T]):
    def __init__(self, name: str, t: type[_T]):
        self._name = __name__ + "." + name
        self._t = t


class HealthService(Service):
    """Tiny HTTP server exposing /healthz for container healthchecks."""

    host: str
    """Host to bind the health server to."""

    port: int
    """Port to bind the health server to."""

    _runner: web.AppRunner | None
    """Aiohttp app runner for the health server."""

    @classmethod
    def create(cls, hassette: "Hassette", host: str = "0.0.0.0", port: int | None = None):
        inst = cls(hassette, parent=hassette)
        inst.host = host
        inst.port = port or hassette.config.health_service_port or 8126
        inst._runner = None

        return inst

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        return self.hassette.config.health_service_log_level

    async def serve(self) -> None:
        if not self.hassette.config.run_health_service:
            return

        try:
            # Just idle until cancelled
            await self.shutdown_event.wait()
        except OSError as e:
            error_no = e.errno if hasattr(e, "errno") else type(e)
            self.logger.error("Health service failed to start: %s (errno=%s)", e, error_no)
            raise

    async def before_initialize(self) -> None:
        self.logger.debug("Waiting for Hassette ready event")
        await self.hassette.ready_event.wait()

    async def on_initialize(self):
        """Start the health HTTP server."""

        if not self.hassette.config.run_health_service:
            self.logger.warning("Health service disabled by configuration")
            # we don't want to fail startup due to "not ready", as this is not unhealthy, just disabled
            self.mark_ready(reason="Health service disabled")
            return

        app = web.Application()
        hassette_key = MyAppKey[HealthService]("health_service", HealthService)
        app[hassette_key] = self
        app.router.add_get("/healthz", self._handle_health)

        self._runner = web.AppRunner(app)
        await self._runner.setup()
        site = web.TCPSite(self._runner, self.host, self.port)
        await site.start()

        self.logger.debug("Health service listening on %s:%s", self.host, self.port)

        self.mark_ready(reason="Health service started")

    async def on_shutdown(self) -> None:
        if self._runner:
            await self._runner.cleanup()
            self.logger.debug("Health service stopped")

    async def _handle_health(self, request: web.Request) -> web.Response:
        # You can check internals here (e.g., WS status)
        ws_running = self.hassette._websocket_service.status == ResourceStatus.RUNNING
        if ws_running:
            self.logger.debug("Health check OK")
            return web.json_response({"status": "ok", "ws": "connected"})
        self.logger.warning("Health check FAILED: WebSocket disconnected")
        return web.json_response({"status": "degraded", "ws": "disconnected"}, status=503)



from copy import deepcopy
from logging import getLogger
from pathlib import Path
from typing import Any
from warnings import warn

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)
from pydantic_settings import BaseSettings
from pydantic_settings.sources import InitSettingsSource, PathType, TomlConfigSettingsSource

DEFAULT_PATH = Path()

LOGGER = getLogger(__name__)


class HassetteTomlConfigSettingsSource(TomlConfigSettingsSource):
    def __init__(self, settings_cls: type[BaseSettings], toml_file: PathType | None = DEFAULT_PATH):
        self.toml_file_path = toml_file if toml_file != DEFAULT_PATH else settings_cls.model_config.get("toml_file")
        self.toml_data = self._read_files(self.toml_file_path)

        if "hassette" not in self.toml_data:
            # just let the standard class handle it
            super().__init__(settings_cls, self.toml_file_path)
            return

        LOGGER.debug("Merging 'hassette' section from TOML config into top level")
        top_level_keys = set(self.toml_data.keys()) - {"hassette"}
        hassette_values = self.toml_data.pop("hassette")
        for key in top_level_keys.intersection(hassette_values.keys()):
            LOGGER.warning(
                "Key %r found in both top level and 'hassette' section of TOML config, "
                "the [hassette] value will be used",
                key,
            )

        self.toml_data.update(hassette_values)

        # need to call InitSettingSource directly, as super() expects a file path
        # as the second argument
        InitSettingsSource.__init__(self, settings_cls, self.toml_data)


class AppManifest(BaseModel):
    """Manifest for a Hassette app."""

    model_config = ConfigDict(
        extra="allow", coerce_numbers_to_str=True, validate_assignment=True, use_attribute_docstrings=True
    )

    app_key: str = Field(default=...)
    """Reflects the key for this app in hassette.toml"""

    enabled: bool = Field(default=True)
    """Whether the app is enabled or not, will default to True if not set"""

    filename: str = Field(default=..., examples=["my_app.py"], validation_alias=AliasChoices("filename", "file_name"))
    """Filename of the app, will be looked for in app_path"""

    class_name: str = Field(
        default=..., examples=["MyApp"], validation_alias=AliasChoices("class_name", "class", "module", "module_name")
    )
    """Class name of the app"""

    display_name: str = Field(default=..., examples=["My App"])
    """Display name of the app, will use class_name if not set"""

    app_dir: Path = Field(..., examples=["./apps"])
    """Path to the app directory, relative to current working directory or absolute"""

    app_config: dict[str, Any] | list[dict[str, Any]] = Field(
        default_factory=dict, validation_alias="config", validate_default=True
    )
    """Instance configuration for the app"""

    auto_loaded: bool = Field(default=False)
    """Whether the app was auto-detected or manually configured"""

    full_path: Path
    """Fully resolved path to the app file"""

    def __repr__(self) -> str:
        return f"<AppManifest {self.display_name} ({self.class_name}) - enabled={self.enabled} file={self.filename}>"

    @model_validator(mode="before")
    @classmethod
    def validate_app_manifest(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate the app configuration."""
        required_keys = ["filename", "class_name", "app_dir"]
        missing_keys = [key for key in required_keys if key not in values]
        if missing_keys:
            raise ValueError(f"App configuration is missing required keys: {', '.join(missing_keys)}")

        values["app_dir"] = app_dir = Path(values["app_dir"]).resolve()

        values["display_name"] = values.get("display_name") or values.get("class_name")

        if app_dir.is_file():
            LOGGER.warning("App directory %s is a file, using the parent directory as app_dir", app_dir)
            values["filename"] = app_dir.name
            values["app_dir"] = app_dir.parent

        return values

    @field_validator("app_config", mode="before")
    @classmethod
    def validate_app_config(cls, v: Any, validation_info: ValidationInfo) -> Any:
        """Set instance name if not set in config."""

        if not v:
            return v

        if isinstance(v, dict):
            v = [v]

        class_name = validation_info.data.get("class_name", "UnknownApp")

        for idx, item in enumerate(v):
            if "instance_name" not in item or not item["instance_name"]:
                item["instance_name"] = f"{class_name}.{idx}"

        return v

    def validate_model_extra(self):
        if not self.model_extra:
            return

        keys = list(self.model_extra.keys())
        msg = (
            f"{type(self).__name__} - {self.display_name} - Instance configuration values should be"
            " set under the `config` field:\n"
            f"  {keys}\n"
            "This will ensure proper validation and handling of custom configurations."
        )

        if not self.app_config:
            self.app_config = deepcopy(self.model_extra)
        elif isinstance(self.app_config, dict) and not set(self.app_config).intersection(set(keys)):
            self.app_config.update(deepcopy(self.model_extra))
        else:
            if isinstance(self.app_config, list):
                msg += "\nNote: Unable to merge extra fields into list-based config."
            elif isinstance(self.app_config, dict):
                msg += "\nNote: Unable to merge extra fields into existing config due to intersecting keys."

            msg += "\nExtra fields will be ignored. Please update your configuration."

        warn(msg, stacklevel=5)

    def model_post_init(self, context: Any) -> None:
        self.validate_model_extra()

        # if we don't have app_config then we don't have any apps with config
        # which means we have, at most, one app
        # so we can just set the default instance name
        if not self.app_config:
            self.app_config = [{"instance_name": f"{self.class_name}.0"}]



from typing import Literal

from pydantic import Field

from hassette.const.sensor import UNIT_OF_MEASUREMENT

from .base import AttributesBase, NumericBaseState


class NumberAttributes(AttributesBase):
    min: int | float | None = Field(default=None)
    max: int | float | None = Field(default=None)
    step: int | float | None = Field(default=None)
    mode: str | None = Field(default=None)
    unit_of_measurement: UNIT_OF_MEASUREMENT | str | None = Field(default=None)


class NumberState(NumericBaseState):
    """Representation of a Home Assistant number state.

    See: https://www.home-assistant.io/integrations/number/
    """

    domain: Literal["number"]

    attributes: NumberAttributes



"""These are quick and dirty fixtures for testing during internal development.

They currently are not meant to be used by external users and will likely not be supported (e.g. bug requests).
However, if you find them useful, knock yourself out.
"""

from .fixtures import (
    hassette_harness,
    hassette_with_app_handler,
    hassette_with_bus,
    hassette_with_file_watcher,
    hassette_with_mock_api,
    hassette_with_scheduler,
    hassette_with_state_proxy,
)
from .harness import HassetteHarness
from .test_server import SimpleTestServer

__all__ = [
    "HassetteHarness",
    "SimpleTestServer",
    "hassette_harness",
    "hassette_with_app_handler",
    "hassette_with_bus",
    "hassette_with_file_watcher",
    "hassette_with_mock_api",
    "hassette_with_scheduler",
    "hassette_with_state_proxy",
]

# TODO: clean these up and make them user facing



import typing

from hassette.bus import Bus
from hassette.events import HassetteServiceEvent
from hassette.resources.base import Resource

if typing.TYPE_CHECKING:
    from hassette import Hassette


class ServiceWatcher(Resource):
    """Watches for service events and handles them."""

    bus: Bus
    """Event bus for inter-service communication."""

    @classmethod
    def create(cls, hassette: "Hassette"):
        inst = cls(hassette, parent=hassette)
        inst.bus = inst.add_child(Bus)
        return inst

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        return self.hassette.config.service_watcher_log_level

    async def on_initialize(self) -> None:
        self._register_internal_event_listeners()
        self.mark_ready(reason="Service watcher initialized")

    async def on_shutdown(self) -> None:
        self.bus.remove_all_listeners()

    async def restart_service(self, event: HassetteServiceEvent) -> None:
        """Start a service from a service event."""
        data = event.payload.data
        name = data.resource_name
        role = data.role

        try:
            if name is None:
                self.logger.warning("No %s specified to start, skipping", role)
                return

            self.logger.debug("%s '%s' is being restarted after '%s'", role, name, event.payload.event_type)

            services = [child for child in self.hassette.children if child.class_name == name and child.role == role]
            if not services:
                self.logger.warning("No %s found for '%s', skipping start", role, name)
                return
            if len(services) > 1:
                self.logger.warning("Multiple %s found for '%s', restarting all", role, name)

            self.logger.debug("Restarting %s '%s'", role, name)
            for service in services:
                await service.restart()

        except Exception as e:
            self.logger.error("Failed to restart %s '%s': %s", role, name, e)
            raise

    async def log_service_event(self, event: HassetteServiceEvent) -> None:
        """Log the startup of a service."""

        name = event.payload.data.resource_name
        role = event.payload.data.role

        if name is None:
            self.logger.warning("No resource specified for startup, cannot log")
            return

        status, previous_status = event.payload.data.status, event.payload.data.previous_status

        if status == previous_status:
            self.logger.debug("%s '%s' status unchanged at '%s', not logging", role, name, status)
            return

        try:
            self.logger.debug(
                "%s '%s' transitioned to status '%s' from '%s'",
                role,
                name,
                event.payload.data.status,
                event.payload.data.previous_status,
            )

        except Exception as e:
            self.logger.error("Failed to log %s startup for '%s': %s", role, name, e)
            raise

    async def shutdown_if_crashed(self, event: HassetteServiceEvent) -> None:
        """Shutdown the Hassette instance if a service has crashed."""
        data = event.payload.data
        name = data.resource_name
        role = data.role

        try:
            self.logger.exception(
                "%s '%s' has crashed (event_id %d), shutting down Hassette, %s",
                role,
                name,
                event.payload.event_id,
                data.exception_traceback,
            )
            await self.hassette.shutdown()
        except Exception:
            self.logger.error("Failed to handle %s crash for '%s': %s", role, name)
            raise

    def _register_internal_event_listeners(self) -> None:
        """Register internal event listeners for resource lifecycle."""
        self.bus.on_hassette_service_failed(handler=self.restart_service)
        self.bus.on_hassette_service_crashed(handler=self.shutdown_if_crashed)
        self.bus.on_hassette_service_status(handler=self.log_service_event)



from pathlib import Path

from watchfiles import awatch

from hassette.events.hassette import HassetteFileWatcherEvent
from hassette.resources.base import Service


class FileWatcherService(Service):
    """Background task to watch for file changes and reload apps."""

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        return self.hassette.config.file_watcher_log_level

    async def before_initialize(self) -> None:
        self.logger.debug("Waiting for Hassette ready event")
        await self.hassette.ready_event.wait()

    async def serve(self) -> None:
        """Watch app directories for changes and trigger reloads."""
        if not self.hassette.config.watch_files:
            self.logger.warning("File watching is disabled due to configuration")
            return

        paths = self.hassette.config.get_watchable_files()

        self.logger.debug("Watching app directories for changes: %s", ", ".join(str(p) for p in paths))
        self.mark_ready(reason="File watcher started")

        async for changes in awatch(
            *paths,
            stop_event=self.shutdown_event,
            step=self.hassette.config.file_watcher_step_milliseconds,
            debounce=self.hassette.config.file_watcher_debounce_milliseconds,
        ):
            if self.shutdown_event.is_set():
                break

            for _, changed_path in changes:
                changed_path = Path(changed_path).resolve()
                self.logger.debug("Detected change in %s", changed_path)
                event = HassetteFileWatcherEvent.create_event(changed_file_path=changed_path)
                await self.hassette.send_event(event.topic, event)

            # update paths in case new apps were added
            paths = self.hassette.config.get_watchable_files()



import typing
from collections.abc import Hashable, Iterator
from dataclasses import dataclass
from logging import getLogger
from typing import ClassVar

from hassette.exceptions import InvalidDataForStateConversionError, InvalidEntityIdError, UnableToConvertStateError
from hassette.utils.exception_utils import get_short_traceback

if typing.TYPE_CHECKING:
    from hassette.events import HassStateDict
    from hassette.models.states.base import BaseState


LOGGER = getLogger(__name__)
CONVERSION_FAIL_TEMPLATE = (
    "Failed to convert state for entity '%s' (domain: '%s') to class '%s'. Data: %s. Error: %s, Traceback: %s"
)


@dataclass(frozen=True)
class StateKey:
    domain: Hashable | None = None
    """The domain of the entity (e.g., 'light', 'sensor')."""

    device_class: Hashable | None = None
    """Optional device class of the entity (e.g., 'temperature', 'humidity'). Not yet being used."""


def register_state_converter(state_class: type["BaseState"], domain: Hashable, device_class: Hashable | None = None):
    """Register a state converter class for a specific domain and optional device class."""
    StateRegistry.register(state_class, domain=domain, device_class=device_class)


class StateRegistry:
    """Registry for mapping domains to their state classes.

    This class maintains a mapping of Home Assistant domains to their corresponding
    BaseState subclasses. State classes get registered during the `after_initialize` phase
    by scanning all subclasses of BaseState.
    """

    _registry: ClassVar[dict[StateKey, type["BaseState"]]] = {}

    def _get_entity_id(self, data: "HassStateDict", entity_id: str | None = None) -> str:
        if not entity_id:
            # specifically this way so we also handle empty strings/None
            entity_id = data.get("entity_id") or "<unknown>"

        if not isinstance(entity_id, str):
            LOGGER.error("State data has invalid 'entity_id' field: %s", data, stacklevel=2)
            raise InvalidEntityIdError(entity_id)

        if "." not in entity_id:
            LOGGER.error("State data has malformed 'entity_id' (missing domain): %s", entity_id, stacklevel=2)
            raise InvalidEntityIdError(entity_id)

        return entity_id

    def try_convert_state(self, data: "HassStateDict", entity_id: str | None = None) -> "BaseState":
        """Convert a dictionary representation of a state into a specific state type.

        This function uses the state registry to look up the appropriate state class
        based on the entity's domain. If no specific class is registered for the domain,
        it falls back to the generic BaseState.

        Args:
            data: Dictionary containing state data from Home Assistant.
            entity_id: Optional entity ID to assist in domain determination.

        Returns:
            A properly typed state object (e.g., LightState, SensorState) or BaseState
            for unknown domains.

        Raises:
            InvalidDataForStateConversionError: If the provided data is invalid or malformed.
            InvalidEntityIdError: If the entity_id is invalid or malformed.
            UnableToConvertStateError: If conversion to the determined state class fails.

        Example:
            ```python
            state_dict = {"entity_id": "light.bedroom", "state": "on", ...}
            light_state = try_convert_state(state_dict)  # Returns LightState instance
            ```
        """
        from hassette.models.states.base import BaseState

        if data is None:
            raise InvalidDataForStateConversionError(data)

        if "event" in data:
            LOGGER.error(
                "Data contains 'event' key, expected state data, not event data. "
                "To convert state from an event, extract the state data from event.payload.data.new_state "
                "or event.payload.data.old_state.",
                stacklevel=2,
            )
            raise InvalidDataForStateConversionError(data)

        entity_id = self._get_entity_id(data, entity_id=entity_id)
        domain = entity_id.split(".", 1)[0]

        # Look up the appropriate state class from the registry
        state_class = self.resolve(domain=domain)

        classes = [state_class, BaseState] if state_class is not None else [BaseState]

        final_idx = len(classes) - 1
        for i, cls in enumerate(classes):
            try:
                return self._conversion_with_error_handling(cls, data, entity_id, domain)
            except UnableToConvertStateError:
                if i == final_idx:
                    raise
                LOGGER.debug(
                    "Falling back to next state class after failure to convert to '%s' for entity '%s'",
                    cls.__name__,
                    entity_id,
                )

        raise RuntimeError("Unreachable code reached in try_convert_state")

    @classmethod
    def register(cls, state_class: type["BaseState"], *, domain: Hashable = None, device_class: Hashable = None):
        key = StateKey(domain=domain, device_class=device_class)
        cls._registry[key] = state_class

    @classmethod
    def resolve(cls, *, domain: Hashable = None, device_class: Hashable = None) -> type["BaseState"] | None:
        """Resolve a state class from the registry based on domain and device_class."""
        candidates = [StateKey(domain=domain, device_class=device_class)]
        if device_class is not None:
            candidates.append(StateKey(domain=domain, device_class=None))

        for k in candidates:
            if k in cls._registry:
                return cls._registry[k]
        return None

    def _conversion_with_error_handling(
        self, state_class: type["BaseState"], data: "HassStateDict", entity_id: str, domain: str
    ) -> "BaseState":
        """Helper to convert state data with error handling.

        This function attempts to convert the given data dictionary into an instance
        of the specified state class. If conversion fails, it logs the error and
        returns None.

        Args:
            state_class: The target state class to convert to.
            data: The state data dictionary.
            entity_id: The entity ID associated with the state data.
            domain: The domain associated with the state data.

        Returns:
            An instance of the state class.

        Raises: UnableToConvertStateError if conversion fails.
        """

        class_name = state_class.__name__
        truncated_data = repr(data)
        if len(truncated_data) > 200:
            truncated_data = truncated_data[:200] + "...[truncated]"

        try:
            return convert_state_dict_to_model(data, state_class)
        except Exception as e:
            tb = get_short_traceback()

            LOGGER.error(
                CONVERSION_FAIL_TEMPLATE,
                entity_id,
                domain,
                class_name,
                truncated_data,
                e,
                tb,
            )
            raise UnableToConvertStateError(entity_id, state_class) from e

    def __contains__(self, model: type["BaseState"]) -> bool:
        """Check if the registry contains a state class for the given model."""
        return any(cls is model for cls in self._registry.values())

    def __iter__(self) -> Iterator[tuple[StateKey, type["BaseState"]]]:
        """Iterate over all registered state classes with their keys."""
        return iter(self._registry.items())

    def items(self) -> Iterator[tuple[StateKey, type["BaseState"]]]:
        """Iterate over all registered state classes with their keys.

        Returns:
            An iterator over (StateKey, BaseState subclass) pairs.
        """
        return iter(self._registry.items())

    def values(self) -> Iterator[type["BaseState"]]:
        """Iterate over all registered state classes.

        Returns:
            An iterator over all registered BaseState subclasses.
        """
        return (state_class for state_class in self._registry.values())

    def keys(self) -> Iterator[StateKey]:
        """Iterate over all registered state keys.

        Returns:
            An iterator over all registered state keys.
        """
        return (key for key in self._registry)


def convert_state_dict_to_model(value: typing.Any, model: type["BaseState"]) -> "BaseState":
    """Convert a raw Home Assistant state dict to a typed state model.

    This converter is used by state object extractors (StateNew, StateOld, etc.) to transform
    the raw state dictionary from Home Assistant into a strongly-typed Pydantic model.

    Args:
        value: The raw state dict from Home Assistant
        model: The target state model class (e.g., LightState, SensorState)

    Returns:
        The typed state model instance

    Raises:
        TypeError: If value is not a dict or model instance
        ValidationError: If the state dict doesn't match the model schema
    """
    if isinstance(value, model):
        return value

    if not isinstance(value, dict):
        raise TypeError(f"Cannot convert {type(value).__name__} to {model.__name__}, expected dict")

    return model.model_validate(value)



import logging

from .api import Api
from .app import App, AppConfig, AppSync, only_app
from .bus import Bus
from .config import HassetteConfig
from .const import ANY_VALUE, MISSING_VALUE, NOT_PROVIDED
from .conversion import (
    STATE_REGISTRY,
    TYPE_REGISTRY,
    TypeConverterEntry,
    register_simple_type_converter,
    register_type_converter_fn,
)
from .core.core import Hassette
from .event_handling import accessors, conditions, dependencies, predicates
from .events import RawStateChangeEvent
from .models import entities, states
from .models.services import ServiceResponse
from .scheduler import Scheduler
from .task_bucket import TaskBucket

A = accessors
C = conditions
D = dependencies
P = predicates

logging.getLogger("hassette").addHandler(logging.NullHandler())

__all__ = [
    "ANY_VALUE",
    "MISSING_VALUE",
    "NOT_PROVIDED",
    "STATE_REGISTRY",
    "TYPE_REGISTRY",
    "A",
    "Api",
    "App",
    "AppConfig",
    "AppSync",
    "Bus",
    "C",
    "D",
    "Hassette",
    "HassetteConfig",
    "P",
    "RawStateChangeEvent",
    "Scheduler",
    "ServiceResponse",
    "TaskBucket",
    "TypeConverterEntry",
    "accessors",
    "conditions",
    "dependencies",
    "entities",
    "only_app",
    "predicates",
    "register_simple_type_converter",
    "register_type_converter_fn",
    "states",
]



import asyncio
import logging
import typing
from logging import getLogger
from typing import Any, ClassVar, Generic, final

from whenever import ZonedDateTime

from hassette.api import Api
from hassette.bus import Bus
from hassette.config.classes import AppManifest
from hassette.events.base import Event
from hassette.resources.base import FinalMeta, Resource
from hassette.scheduler import Scheduler
from hassette.state_manager import StateManager
from hassette.types import AppConfigT
from hassette.types.enums import ResourceRole
from hassette.utils.date_utils import now

from .app_config import AppConfig

if typing.TYPE_CHECKING:
    from hassette import Hassette

LOGGER = getLogger(__name__)

AppT = typing.TypeVar("AppT", bound="App")


def only_app(app_cls: type[AppT]) -> type[AppT]:
    """Decorator to mark an app class as the only one to run. If more than one app is marked with this decorator,
    an exception will be raised during initialization.

    This is useful for development and testing, where you may want to run only a specific app without
    modifying configuration files.
    """
    app_cls._only_app = True
    return app_cls


class App(Generic[AppConfigT], Resource, metaclass=FinalMeta):
    """Base class for applications in the Hassette framework.

    This class provides a structure for applications, allowing them to be initialized and managed
    within the Hassette ecosystem. Lifecycle will generally be managed for you via the service status events,
    which send an event to the Bus and set the `status` attribute, based on the app's lifecycle.
    """

    _only_app: ClassVar[bool] = False
    """If True, only this app will be run. Only one app can be marked as only."""

    _import_exception: ClassVar[Exception | None] = None
    """Exception raised during import, if any. This prevents having all apps in a module fail due to one exception."""

    role: ClassVar[ResourceRole] = ResourceRole.APP
    """Role of the resource, e.g. 'App', 'Service', etc."""

    app_manifest: ClassVar[AppManifest]
    "Manifest for the app itself, not used by app instances."

    app_config_cls: ClassVar[type[AppConfig]]
    """Config class to use for instances of the created app. Configuration from hassette.toml or
    other sources will be validated by this class."""

    logger: logging.Logger
    """Logger for the instance."""

    api: "Api"
    """API instance for interacting with Home Assistant."""

    scheduler: "Scheduler"
    """Scheduler instance for scheduled tasks owned by this app."""

    bus: "Bus"
    """Event bus instance for event handlers owned by this app."""

    states: "StateManager"
    """States manager instance for accessing Home Assistant states."""

    app_config: AppConfigT
    """Configuration for this app instance."""

    index: int
    """Index of this app instance, used for unique naming."""

    def __init__(self, *args, app_config: AppConfigT, index: int, **kwargs):
        # unlike most classes, this one does take additional init args
        # this is because the unique name we use for the logger depends on the app config
        self.app_config = app_config
        self.index = index
        super().__init__(*args, **kwargs)

    @classmethod
    def create(cls, hassette: "Hassette", app_config: AppConfigT, index: int):
        inst = cls(hassette=hassette, app_config=app_config, index=index)
        inst.app_config = app_config
        inst.index = index
        inst.api = inst.add_child(Api)
        inst.scheduler = inst.add_child(Scheduler)
        inst.bus = inst.add_child(Bus, priority=0)
        inst.states = inst.add_child(StateManager)
        return inst

    @property
    def unique_name(self) -> str:
        """Unique name for the app instance, used for logging and ownership of resources."""
        if self.app_config.instance_name.startswith(self.class_name):
            return self.app_config.instance_name
        return f"{self.class_name}.{self.app_config.instance_name}"

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        try:
            return self.app_config.log_level
        except AttributeError:
            return self.hassette.config.apps_log_level

    @property
    def instance_name(self) -> str:
        """Name for the instance of the app. Used for logging and ownership of resources."""
        return self.app_config.instance_name

    def now(self) -> ZonedDateTime:
        """Return the current date and time."""
        return now()

    async def send_event(self, event_name: str, event: Event[Any]) -> None:
        """Send an event to the event bus."""
        await self.hassette.send_event(event_name, event)

    @final
    async def cleanup(self, timeout: int | None = None) -> None:
        """Cleanup resources owned by the instance.

        This method is called during shutdown to ensure that all resources are properly released.
        """
        timeout = timeout or self.hassette.config.app_shutdown_timeout_seconds

        await super().cleanup(timeout=timeout)

        tasks = []

        tasks.append(self.scheduler.remove_all_jobs())
        tasks.append(self.bus.remove_all_listeners())
        tasks.append(self.task_bucket.cancel_all())

        if tasks:
            results = await asyncio.wait_for(asyncio.gather(*tasks, return_exceptions=True), timeout=timeout)
            for result in results:
                if isinstance(result, Exception):
                    self.logger.error("Error during resource cleanup for app '%s': %s", self.class_name, result)
        self.logger.debug("All resources cleaned up for app '%s'", self.class_name)


class AppSync(App[AppConfigT]):
    """Synchronous adapter for App."""

    def send_event_sync(self, event_name: str, event: Event[Any]) -> None:
        """Synchronous version of send_event."""
        self.task_bucket.run_sync(self.send_event(event_name, event))

    @final
    async def before_shutdown(self) -> None:
        """Optional: stop accepting new work, signal loops to wind down, etc."""
        await self.task_bucket.run_in_thread(self.before_shutdown_sync)

    @final
    async def on_shutdown(self) -> None:
        """Primary hook: release your own stuff (sockets, queues, temp files…)."""
        await self.task_bucket.run_in_thread(self.on_shutdown_sync)

    @final
    async def after_shutdown(self) -> None:
        """Optional: last-chance actions after on_shutdown, before cleanup/STOPPED."""
        await self.task_bucket.run_in_thread(self.after_shutdown_sync)

    @final
    async def before_initialize(self) -> None:
        """Optional: prepare to accept new work, allocate sockets, queues, temp files, etc."""
        await self.task_bucket.run_in_thread(self.before_initialize_sync)

    @final
    async def on_initialize(self) -> None:
        """Primary hook: perform your own initialization (sockets, queues, temp files…)."""
        await self.task_bucket.run_in_thread(self.on_initialize_sync)

    @final
    async def after_initialize(self) -> None:
        """Optional: finalize initialization, signal readiness, etc."""
        await self.task_bucket.run_in_thread(self.after_initialize_sync)

    def before_shutdown_sync(self) -> None:
        """Optional: stop accepting new work, signal loops to wind down, etc."""
        pass

    def on_shutdown_sync(self) -> None:
        """Primary hook: release your own stuff (sockets, queues, temp files…)."""
        pass

    def after_shutdown_sync(self) -> None:
        """Optional: last-chance actions after on_shutdown, before cleanup/STOPPED."""
        pass

    def before_initialize_sync(self) -> None:
        """Optional: prepare to accept new work, allocate sockets, queues, temp files, etc."""
        pass

    def on_initialize_sync(self) -> None:
        """Primary hook: perform your own initialization (sockets, queues, temp files…)."""
        pass

    def after_initialize_sync(self) -> None:
        """Optional: finalize initialization, signal readiness, etc."""
        pass

    @final
    def initialize_sync(self) -> None:
        """Use on_initialize_sync instead."""
        raise NotImplementedError("Use on_initialize_sync instead.")

    @final
    def shutdown_sync(self) -> None:
        """Use on_shutdown_sync instead."""
        raise NotImplementedError("Use on_shutdown_sync instead.")



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class ClimateAttributes(AttributesBase):
    hvac_modes: list[str] | None = Field(default=None)
    min_temp: int | float | None = Field(default=None)
    max_temp: int | float | None = Field(default=None)
    fan_modes: list[str] | None = Field(default=None)
    preset_modes: list[str] | None = Field(default=None)
    current_temperature: int | float | None = Field(default=None)
    temperature: int | float | None = Field(default=None)
    target_temp_high: float | None = Field(default=None)
    target_temp_low: float | None = Field(default=None)
    current_humidity: float | None = Field(default=None)
    fan_mode: str | None = Field(default=None)
    hvac_action: str | None = Field(default=None)
    preset_mode: str | None = Field(default=None)
    swing_mode: str | None = Field(default=None)
    swing_modes: list[str] | None = Field(default=None)


class ClimateState(StringBaseState):
    """Representation of a Home Assistant climate state.

    See: https://www.home-assistant.io/integrations/climate/
    """

    domain: Literal["climate"]

    attributes: ClimateAttributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class FanAttributes(AttributesBase):
    preset_modes: list[str] | None = Field(default=None)
    oscillating: bool | None = Field(default=None)
    percentage: int | float | None = Field(default=None)
    percentage_step: float | None = Field(default=None)
    preset_mode: str | None = Field(default=None)
    temperature: int | float | None = Field(default=None)
    model: str | None = Field(default=None)
    sn: str | None = Field(default=None)
    screen_status: bool | None = Field(default=None)
    child_lock: bool | None = Field(default=None)
    night_light: str | None = Field(default=None)
    mode: str | None = Field(default=None)


class FanState(StringBaseState):
    """Representation of a Home Assistant fan state.

    See: https://www.home-assistant.io/integrations/fan/
    """

    domain: Literal["fan"]

    attributes: FanAttributes



from typing import Literal

from pydantic import Field

from .base import AttributesBase, DateTimeBaseState


class EventAttributes(AttributesBase):
    event_types: list[str] | None = Field(default=None)
    event_type: str | None = Field(default=None)
    button: str | None = Field(default=None)


class EventState(DateTimeBaseState):
    """Representation of a Home Assistant event state.

    See: https://www.home-assistant.io/integrations/event/
    """

    domain: Literal["event"]

    attributes: EventAttributes



import asyncio
import typing
from collections import defaultdict
from copy import deepcopy
from logging import getLogger
from pathlib import Path
from timeit import default_timer as timer

import anyio
from deepdiff import DeepDiff
from humanize import precisedelta

import hassette.event_handling.accessors as A
from hassette.app.app import App
from hassette.bus import Bus
from hassette.events.hassette import HassetteSimpleEvent
from hassette.exceptions import InvalidInheritanceError, UndefinedUserConfigError
from hassette.resources.base import Resource
from hassette.types import Topic
from hassette.types.enums import ResourceStatus
from hassette.utils.app_utils import (
    class_already_loaded,
    class_failed_to_load,
    get_class_load_error,
    get_loaded_class,
    load_app_class_from_manifest,
)
from hassette.utils.exception_utils import get_short_traceback

if typing.TYPE_CHECKING:
    from hassette import AppConfig, Hassette
    from hassette.config.classes import AppManifest

LOGGER = getLogger(__name__)
LOADED_CLASSES: "dict[tuple[str, str], type[App[AppConfig]]]" = {}
ROOT_PATH = "root"
USER_CONFIG_PATH = "user_config"


class AppHandler(Resource):
    """Manages the lifecycle of apps in Hassette.

    - Deterministic storage: apps[app_name][index] -> App
    - Tracks per-app failures in failed_apps for observability
    """

    # TODO:
    # need to separate startup of app handler from initialization of apps
    # so that we can start the app handler, then the API, then initialize apps
    # because apps may want to use the API during startup
    # could trigger on websocket connected event, with a once=True handler?

    # TODO: handle stopping/starting individual app instances, instead of all apps of a class/key
    # no need to restart app index 2 if only app index 0 changed, etc.

    # TODO: clean this class up - it likely needs to be split into smaller pieces

    apps_config: dict[str, "AppManifest"]
    """Copy of Hassette's config apps"""

    apps: dict[str, dict[int, App["AppConfig"]]]
    """Running apps"""

    failed_apps: dict[str, list[tuple[int, Exception]]]
    """Apps we could not start/failed to start"""

    only_app: str | None
    """If set, only this app will be started (the one marked as only)"""

    bus: Bus
    """Event bus for inter-service communication."""

    @classmethod
    def create(cls, hassette: "Hassette"):
        inst = cls(hassette, parent=hassette)
        inst.apps_config = {}
        inst.set_apps_configs(hassette.config.app_manifests)
        inst.only_app = None
        inst.apps = defaultdict(dict)
        inst.failed_apps = defaultdict(list)
        inst.bus = inst.add_child(Bus)
        return inst

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        return self.hassette.config.app_handler_log_level

    def set_apps_configs(self, apps_config: dict[str, "AppManifest"]) -> None:
        """Set the apps configuration.

        Args:
            apps_config: The new apps configuration.
        """
        self.logger.debug("Setting apps configuration")
        self.apps_config = deepcopy(apps_config)
        self.only_app = None  # reset only_app, will be recomputed on next initialize

        self.logger.debug("Found %d apps in configuration: %s", len(self.apps_config), list(self.apps_config.keys()))

    @property
    def active_apps_config(self) -> dict[str, "AppManifest"]:
        """Apps that are enabled."""
        enabled_apps = {k: v for k, v in self.apps_config.items() if v.enabled}
        if self.only_app:
            enabled_apps = {k: v for k, v in enabled_apps.items() if k == self.only_app}
        return enabled_apps

    async def on_initialize(self) -> None:
        """Start handler and initialize configured apps."""
        if self.hassette.config.dev_mode or self.hassette.config.allow_reload_in_prod:
            if self.hassette.config.allow_reload_in_prod:
                self.logger.warning("Allowing app reloads in production mode due to config")
            self.bus.on(topic=Topic.HASSETTE_EVENT_FILE_WATCHER, handler=self.handle_change_event)
        else:
            self.logger.debug("Not watching for app changes, dev_mode is disabled")

        await self.hassette.wait_for_ready(self.hassette._websocket_service)
        self.mark_ready("initialized")

    async def after_initialize(self) -> None:
        self.logger.debug("Scheduling app initialization")
        self.task_bucket.spawn(self.initialize_apps())

    async def on_shutdown(self) -> None:
        """Shutdown all app instances gracefully."""
        self.logger.debug("Stopping '%s' %s", self.class_name, self.role)
        self.mark_not_ready(reason="shutting-down")

        self.bus.remove_all_listeners()

        # Flatten and iterate
        for instances in list(self.apps.values()):
            for inst in list(instances.values()):
                try:
                    with anyio.fail_after(self.hassette.config.app_shutdown_timeout_seconds):
                        await inst.shutdown()

                        # in case the app does not call its own cleanup
                        # which is honestly a better user experience
                        await inst.cleanup()
                    self.logger.debug("App %s shutdown successfully", inst.app_config.instance_name)
                except Exception:
                    self.logger.error(
                        "Failed to shutdown app %s:\n%s", inst.app_config.instance_name, get_short_traceback()
                    )

        self.apps.clear()
        self.failed_apps.clear()

    def get(self, app_key: str, index: int = 0) -> "App[AppConfig] | None":
        """Get a specific app instance if running."""
        return self.apps.get(app_key, {}).get(index)

    def all(self) -> list["App[AppConfig]"]:
        """All running app instances."""
        return [inst for group in self.apps.values() for inst in group.values()]

    async def initialize_apps(self) -> None:
        """Initialize all configured and enabled apps, called at AppHandler startup."""

        if not self.apps_config:
            self.logger.debug("No apps configured, skipping initialization")
            return

        if not await self.hassette.wait_for_ready(
            [
                self.hassette._websocket_service,
                self.hassette._api_service,
                self.hassette._bus_service,
                self.hassette._scheduler_service,
                self.hassette._state_proxy,
            ]
        ):
            self.logger.warning("Dependencies never became ready; skipping app startup")
            return

        try:
            tasks = await self._initialize_apps()
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    self.logger.exception("Error during app initialization: %s", result)
            if not self.apps:
                self.logger.warning("No apps were initialized successfully")
            else:
                success_count = sum(
                    len([a for a in v.values() if a.status == ResourceStatus.RUNNING]) for v in self.apps.values()
                )
                fail_count = sum(len(v) for v in self.failed_apps.values())
                self.logger.debug("Initialized %d apps successfully, %d failed to start", success_count, fail_count)

            await self.hassette.send_event(
                Topic.HASSETTE_EVENT_APP_LOAD_COMPLETED,
                HassetteSimpleEvent.create_event(topic=Topic.HASSETTE_EVENT_APP_LOAD_COMPLETED),
            )
        except Exception as e:
            self.logger.exception("Failed to initialize apps")
            await self.handle_crash(e)
            raise

    async def _set_only_app(self):
        """Determine if any app is marked as only, and set self.only_app accordingly."""

        only_apps: list[str] = []
        for app_manifest in self.active_apps_config.values():
            try:
                if class_failed_to_load(app_manifest.full_path, app_manifest.class_name):
                    self.logger.debug(
                        "Skipping only_app check for '%s' because class failed to load", app_manifest.app_key
                    )
                    continue
                if class_already_loaded(app_manifest.full_path, app_manifest.class_name):
                    app_class = get_loaded_class(app_manifest.full_path, app_manifest.class_name)
                else:
                    app_class = load_app_class_from_manifest(app_manifest)
                if app_class._only_app:
                    only_apps.append(app_manifest.app_key)
            except (UndefinedUserConfigError, InvalidInheritanceError):
                self.logger.error(
                    "Failed to load app '%s' due to bad configuration - check previous logs for details",
                    app_manifest.display_name,
                )
            except Exception:
                self.logger.error(
                    "Failed to load app class for '%s':\n%s", app_manifest.display_name, get_short_traceback()
                )

        if not only_apps:
            self.only_app = None
            return

        if not self.hassette.config.dev_mode:
            if not self.hassette.config.allow_only_app_in_prod:
                self.logger.warning("Disallowing use of `only_app` decorator in production mode")
                self.only_app = None
                return
            self.logger.warning("Allowing use of `only_app` decorator in production mode due to config")

        if len(only_apps) > 1:
            keys = ", ".join(app for app in only_apps)
            raise RuntimeError(f"Multiple apps marked as only: {keys}")

        self.only_app = only_apps[0]
        self.logger.warning("App %s is marked as only, skipping all others", self.only_app)

    async def _initialize_apps(self, apps: set[str] | None = None) -> list[asyncio.Task]:
        """Initialize all or a subset of apps by key. If apps is None, initialize all enabled apps."""

        tasks: list[asyncio.Task] = []
        await self._set_only_app()

        apps = apps if apps is not None else set(self.active_apps_config.keys())

        for app_key in apps:
            app_manifest = self.active_apps_config.get(app_key)
            if not app_manifest:
                self.logger.debug("Skipping disabled or unknown app %s", app_key)
                continue
            try:
                self._create_app_instances(app_key, app_manifest)
            except (UndefinedUserConfigError, InvalidInheritanceError):
                self.logger.error(
                    "Failed to load app '%s' due to bad configuration - check previous logs for details", app_key
                )
                continue
            except Exception:
                self.logger.error("Failed to load app class for '%s':\n%s", app_key, get_short_traceback())
                continue

            tasks.append(self.task_bucket.spawn(self._initialize_app_instances(app_key, app_manifest)))

        return tasks

    def _create_app_instances(self, app_key: str, app_manifest: "AppManifest", force_reload: bool = False) -> None:
        """Create app instances from a manifest, validating config.

        Args:
            app_key: The key of the app, as found in hassette.toml.
            app_manifest: The manifest containing configuration.
        """

        already_loaded = class_already_loaded(app_manifest.full_path, app_manifest.class_name)
        already_failed = class_failed_to_load(app_manifest.full_path, app_manifest.class_name)

        # if we are forcing a reload we have to try again
        # or if we've never loaded it before/failed to load it before, we have to try to load it
        if force_reload or (not already_loaded and not already_failed):
            try:
                app_class = load_app_class_from_manifest(app_manifest, force_reload=force_reload)
            except Exception as e:
                self.logger.error("Failed to load app class for '%s':\n%s", app_key, get_short_traceback())
                self.failed_apps[app_key].append((0, e))
                return
        # if we've already failed to load it (and we're not forcing a reload), we can't load it again
        elif already_failed:
            self.logger.debug("Cannot create app instances for '%s' because class failed to load previously", app_key)
            load_error = get_class_load_error(app_manifest.full_path, app_manifest.class_name)
            self.failed_apps[app_key].append((0, load_error))
            return
        # else, just get the already loaded class
        elif already_loaded:
            app_class = get_loaded_class(app_manifest.full_path, app_manifest.class_name)

        class_name = app_class.__name__
        app_class.app_manifest = app_manifest
        app_configs = app_manifest.app_config

        # toml data can be a dict or a list of dicts
        # AppManifest should handle conversion for us but we're being cautious here
        app_configs = [app_configs] if not isinstance(app_configs, list) else app_configs

        for idx, config in enumerate(app_configs):
            instance_name = config.get("instance_name")
            if not instance_name:
                raise ValueError(f"App {app_key} instance {idx} is missing instance_name")
            try:
                validated = app_class.app_config_cls.model_validate(config)
                app_instance = app_class.create(hassette=self.hassette, app_config=validated, index=idx)
                self.apps[app_key][idx] = app_instance
            except Exception as e:
                self.logger.error(
                    "Failed to validate/init config for %s (%s):\n%s",
                    instance_name,
                    class_name,
                    get_short_traceback(),
                )
                self.failed_apps[app_key].append((idx, e))
                continue

    async def _initialize_app_instances(self, app_key: str, app_manifest: "AppManifest") -> None:
        """Initialize all instances of a given app_key.

        Args:
            app_key: The key of the app, as found in hassette.toml.
            app_manifest (AppManifest): The manifest containing configuration.

        """

        class_name = app_manifest.class_name
        for idx, inst in self.apps.get(app_key, {}).items():
            try:
                with anyio.fail_after(self.hassette.config.app_startup_timeout_seconds):
                    await inst.initialize()
                    inst.mark_ready(reason="initialized")
                self.logger.debug("App '%s' (%s) initialized successfully", inst.app_config.instance_name, class_name)
            except TimeoutError as e:
                self.logger.error(
                    "Timed out while starting app '%s' (%s):\n%s",
                    inst.app_config.instance_name,
                    class_name,
                    get_short_traceback(5),
                )
                inst.status = ResourceStatus.STOPPED
                self.failed_apps[app_key].append((idx, e))
            except Exception as e:
                self.logger.error(
                    "Failed to start app '%s' (%s):\n%s",
                    inst.app_config.instance_name,
                    class_name,
                    get_short_traceback(5),
                )
                inst.status = ResourceStatus.STOPPED
                self.failed_apps[app_key].append((idx, e))

    async def refresh_config(self) -> tuple[dict[str, "AppManifest"], dict[str, "AppManifest"]]:
        """Reload the configuration and return (original_apps_config, current_apps_config)."""
        original_apps_config = deepcopy(self.active_apps_config)

        # Reinitialize config to pick up changes.
        # https://docs.pydantic.dev/latest/concepts/pydantic_settings/#in-place-reloading
        try:
            self.hassette.config.reload()
        except Exception as e:
            self.logger.exception("Failed to reload configuration: %s", e)

        self.set_apps_configs(self.hassette.config.app_manifests)
        curr_apps_config = deepcopy(self.active_apps_config)

        return original_apps_config, curr_apps_config

    async def handle_change_event(
        self,
        changed_file_path: typing.Annotated[Path | None, A.get_path("payload.data.changed_file_path")] = None,
    ) -> None:
        """Handle changes detected by the watcher."""

        original_apps_config, curr_apps_config = await self.refresh_config()

        # recalculate only_app in case it changed
        await self._set_only_app()

        orphans, new_apps, reimport_apps, reload_apps = self._calculate_app_changes(
            original_apps_config, curr_apps_config, changed_file_path
        )
        self.logger.debug(
            "App changes detected - orphans: %s, new: %s, reimport: %s, reload: %s",
            orphans,
            new_apps,
            reimport_apps,
            reload_apps,
        )
        await self._handle_removed_apps(orphans)
        await self._handle_new_apps(new_apps)
        await self._reload_apps_due_to_file_change(reimport_apps)
        await self._reload_apps_due_to_config(reload_apps)

        await self.hassette.send_event(
            Topic.HASSETTE_EVENT_APP_LOAD_COMPLETED,
            HassetteSimpleEvent.create_event(topic=Topic.HASSETTE_EVENT_APP_LOAD_COMPLETED),
        )

    def _calculate_app_changes(
        self,
        original_apps_config: dict[str, "AppManifest"],
        curr_apps_config: dict[str, "AppManifest"],
        changed_path: Path | None,
    ) -> tuple[set[str], set[str], set[str], set[str]]:
        """Return 4 sets of app keys: (orphans, new_apps, reimport_apps, reload_apps).

        Args:
            original_apps_config: The original apps configuration.
            curr_apps_config: The current apps configuration.
            changed_path: The path of the file that changed, if any.

        Returns:
            A tuple containing four sets:
                - orphans: Apps that were removed from the configuration.
                - new_apps: Apps that were added to the configuration.
                - reimport_apps: Apps that need to be reimported due to file changes.
                - reload_apps: Apps that need to be reloaded due to configuration changes.
        """

        config_diff = DeepDiff(
            original_apps_config, curr_apps_config, ignore_order=True, include_paths=[ROOT_PATH, USER_CONFIG_PATH]
        )

        original_app_keys = set(original_apps_config.keys())
        curr_app_keys = set(curr_apps_config.keys())
        if self.only_app:
            curr_app_keys = {k for k in curr_app_keys if k == self.only_app}

        orphans = original_app_keys - curr_app_keys
        new_apps = curr_app_keys - original_app_keys

        reimport_apps = {app.app_key for app in curr_apps_config.values() if app.full_path == changed_path}

        reload_apps = {
            app_key
            for app_key in config_diff.affected_root_keys
            if app_key not in new_apps and app_key not in orphans and app_key not in reimport_apps
        }

        return orphans, new_apps, reimport_apps, reload_apps

    async def _handle_removed_apps(self, orphans: set[str]) -> None:
        if not orphans:
            return

        self.logger.debug("Apps removed from config: %s", orphans)

        self.logger.debug("Stopping %d orphaned apps: %s", len(orphans), orphans)
        for app_key in orphans:
            self.logger.debug("Stopping orphaned app %s", app_key)
            try:
                await self.stop_app(app_key)
            except Exception:
                self.logger.error("Failed to stop orphaned app %s:\n%s", app_key, get_short_traceback())

    async def _reload_apps_due_to_file_change(self, apps: set[str]) -> None:
        if not apps:
            return

        self.logger.debug("Apps to reimport due to file change: %s", apps)
        for app_key in apps:
            await self.reload_app(app_key, force_reload=True)

    async def _reload_apps_due_to_config(self, apps: set[str]) -> None:
        if not apps:
            return

        self.logger.debug("Apps to reload due to config changes: %s", apps)
        for app_key in apps:
            await self.reload_app(app_key)

    async def stop_app(self, app_key: str) -> None:
        """Stop and remove all instances for a given app_name."""
        instances = self.apps.pop(app_key, None)
        if not instances:
            self.logger.warning("Cannot stop app %s, not found", app_key)
            return
        self.logger.debug("Stopping %d instances of %s", len(instances), app_key)

        for inst in instances.values():
            try:
                start_time = timer()
                with anyio.fail_after(self.hassette.config.app_shutdown_timeout_seconds):
                    await inst.shutdown()

                end_time = timer()
                friendly_time = precisedelta(end_time - start_time, minimum_unit="milliseconds")
                self.logger.debug("Stopped app '%s' in %s", inst.app_config.instance_name, friendly_time)

            except Exception:
                self.logger.error(
                    "Failed to stop app '%s' after %s seconds:\n%s",
                    inst.app_config.instance_name,
                    self.hassette.config.app_shutdown_timeout_seconds,
                    get_short_traceback(),
                )

    async def _handle_new_apps(self, apps: set[str]) -> None:
        """Start any apps that are in config but not currently running."""
        if not apps:
            return

        self.logger.debug("Starting %d new apps: %s", len(apps), list(apps))
        await self._initialize_apps(apps)

    async def reload_app(self, app_key: str, force_reload: bool = False) -> None:
        """Stop and reinitialize a single app by key (based on current config)."""
        self.logger.debug("Reloading app %s", app_key)
        try:
            await self.stop_app(app_key)
            # Initialize only that app from the current config if present and enabled
            manifest = self.active_apps_config.get(app_key)
            if not manifest:
                if manifest := self.apps_config.get(app_key):
                    self.logger.warning("Cannot reload app %s, not enabled", app_key)
                    return
                self.logger.warning("Cannot reload app %s, not found", app_key)
                return

            assert manifest is not None, "Manifest should not be None"

            self._create_app_instances(app_key, manifest, force_reload=force_reload)
            await self._initialize_app_instances(app_key, manifest)
        except Exception:
            self.logger.error("Failed to reload app %s:\n%s", app_key, get_short_traceback())



import contextlib
import inspect
import itertools
import typing
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from logging import Logger, getLogger
from typing import Any, cast

from hassette.bus.injection import ParameterInjector
from hassette.bus.rate_limiter import RateLimiter
from hassette.event_handling.predicates import normalize_where
from hassette.utils.func_utils import callable_name, callable_short_name
from hassette.utils.type_utils import get_typed_signature

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from hassette import TaskBucket
    from hassette.events.base import Event
    from hassette.types import AsyncHandlerType, HandlerType, Predicate

LOGGER = getLogger(__name__)

seq = itertools.count(1)


def next_id() -> int:
    return next(seq)


@dataclass(slots=True)
class Listener:
    """A listener for events with a specific topic and handler."""

    logger: Logger
    """Logger for the listener."""

    listener_id: int = field(default_factory=next_id, init=False)
    """Unique identifier for the listener instance."""

    owner: str
    """Unique string identifier for the owner of the listener, e.g., a component or integration name."""

    topic: str
    """Topic the listener is subscribed to."""

    orig_handler: "HandlerType"
    """Original handler function provided by the user."""

    adapter: "HandlerAdapter"
    """Handler adapter that manages signature normalization and rate limiting."""

    predicate: "Predicate | None"
    """Predicate to filter events before invoking the handler."""

    kwargs: Mapping[str, Any] | None = None
    """Keyword arguments to pass to the handler."""

    once: bool = False
    """Whether the listener should be removed after one invocation."""

    priority: int = 0
    """Priority for listener ordering. Higher values run first. Default is 0 for app handlers."""

    @property
    def handler_name(self) -> str:
        return callable_name(self.orig_handler)

    @property
    def handler_short_name(self) -> str:
        return callable_short_name(self.orig_handler)

    async def matches(self, ev: "Event[Any]") -> bool:
        """Check if the event matches the listener's predicate."""
        if self.predicate is None:
            return True
        matched = self.predicate(ev)

        match_str = "matched" if matched else "did not match"
        self.logger.debug("Listener %s %s predicate for event: %s", self, match_str, ev)
        return matched

    async def invoke(self, event: "Event[Any]") -> None:
        """Invoke the handler through the adapter."""
        kwargs = self.kwargs or {}
        await self.adapter.call(event, **kwargs)

    def __repr__(self) -> str:
        return f"Listener<{self.owner} - {self.handler_short_name}>"

    @classmethod
    def create(
        cls,
        task_bucket: "TaskBucket",
        owner: str,
        topic: str,
        handler: "HandlerType",
        where: "Predicate | Sequence[Predicate] | None" = None,
        kwargs: Mapping[str, Any] | None = None,
        once: bool = False,
        debounce: float | None = None,
        throttle: float | None = None,
        priority: int = 0,
        logger: Logger = LOGGER,
    ) -> "Listener":
        pred = normalize_where(where)
        signature = get_typed_signature(handler)

        # Create async handler
        async_handler = make_async_handler(handler, task_bucket)

        # Create an adapter with rate limiting and signature informed calling
        adapter = HandlerAdapter(
            callable_name(handler),
            async_handler,
            signature,
            task_bucket,
            debounce=debounce,
            throttle=throttle,
        )

        return cls(
            logger=logger,
            owner=owner,
            topic=topic,
            orig_handler=handler,
            adapter=adapter,
            predicate=pred,
            kwargs=kwargs,
            once=once,
            priority=priority,
        )


class HandlerAdapter:
    """Handler adapter that composes dependency injection and rate limiting."""

    def __init__(
        self,
        handler_name: str,
        handler: "AsyncHandlerType",
        signature: inspect.Signature,
        task_bucket: "TaskBucket",
        debounce: float | None = None,
        throttle: float | None = None,
    ):
        self.handler_name = handler_name
        self.handler = handler
        self.task_bucket = task_bucket

        # Dependency injection setup
        self.injector = ParameterInjector(handler_name, signature)

        # Rate limiting setup
        self.rate_limiter: RateLimiter | None = None
        if debounce or throttle:
            if debounce and throttle:
                raise ValueError("Cannot specify both 'debounce' and 'throttle' parameters")
            self.rate_limiter = RateLimiter(
                task_bucket=task_bucket,
                debounce=debounce,
                throttle=throttle,
            )

    async def call(self, event: "Event[Any]", **kwargs: Any) -> None:
        """Call handler with dependency injection and optional rate limiting.

        Args:
            event: The event to pass to the handler.
            **kwargs: Additional keyword arguments to pass to the handler.

        Raises:
            DependencyInjectionError: If signature validation fails.
            DependencyResolutionError: If parameter extraction/conversion fails.
            Exception: If an error occurs during handler execution.
        """
        if self.rate_limiter:
            await self.rate_limiter.call(self._direct_call, event, **kwargs)
        else:
            await self._direct_call(event, **kwargs)

    async def _direct_call(self, event: "Event[Any]", **kwargs: Any) -> None:
        """Call handler directly with dependency injection (no rate limiting).

        Args:
            event: The event to pass to the handler.
            **kwargs: Additional keyword arguments to pass to the handler.
        """
        # Inject parameters
        kwargs = self.injector.inject_parameters(event, **kwargs)

        await self.handler(**kwargs)


@dataclass(slots=True)
class Subscription:
    """A subscription to an event topic with a specific listener key.

    This class is used to manage the lifecycle of a listener, allowing it to be cancelled
    or managed within a context.
    """

    listener: Listener
    """The listener associated with this subscription."""

    unsubscribe: "Callable[[], None]"
    """Function to call to unsubscribe the listener."""

    @contextlib.contextmanager
    def manage(self):
        try:
            yield self
        finally:
            self.unsubscribe()

    def cancel(self) -> None:
        """Cancel the subscription by calling the unsubscribe function."""
        self.unsubscribe()


def make_async_handler(fn: "HandlerType", task_bucket: "TaskBucket") -> "AsyncHandlerType":
    """Wrap a function to ensure it is always called as an async handler.

    If the function is already an async function, it will be called directly.
    If it is a regular function, it will be run in an executor to avoid blocking the event loop.

    Args:
        fn: The function to adapt.

    Returns:
        An async handler that wraps the original function.
    """
    return cast("AsyncHandlerType", task_bucket.make_async_adapter(fn))



from typing import Literal

from pydantic import Field

from hassette.const.sensor import DEVICE_CLASS, STATE_CLASS, UNIT_OF_MEASUREMENT

from .base import AttributesBase, StringBaseState


class SensorAttributes(AttributesBase):
    device_class: DEVICE_CLASS | str | None = Field(default=None)

    state_class: STATE_CLASS | str | None = Field(default=None)
    unit_of_measurement: UNIT_OF_MEASUREMENT | str | None = Field(default=None)

    options: list[str] | None = Field(default=None)


class SensorState(StringBaseState):
    """Representation of a Home Assistant sensor state.

    See: https://www.home-assistant.io/integrations/sensor/"""

    domain: Literal["sensor"]
    attributes: SensorAttributes



import traceback
from textwrap import indent


def get_traceback_string(exception: Exception | BaseException) -> str:
    """Get a formatted traceback string from an exception."""

    return "".join(traceback.format_exception(type(exception), exception, exception.__traceback__))


def get_short_traceback(limit: int = 1, indent_output: bool = True) -> str:
    """Get a short traceback string of the current exception. Must be called within an exception handler.

    Args:
        limit: The number of stack levels to include in the traceback.
        indent_output: Whether to indent the output for better readability.

    Returns:
        A formatted traceback string.

    """
    value = traceback.format_exc(limit=limit)
    if indent_output:
        value = indent(value, "    ")

    lines = value.splitlines()
    if lines and not lines[-1].strip():
        lines = lines[:-1]

    value = "\n".join(lines)
    return value



import asyncio
import typing

if typing.TYPE_CHECKING:
    from hassette.resources.base import Resource


async def wait_for_ready(
    resources: "list[Resource] | Resource",
    poll_interval: float = 0.1,
    timeout: int = 20,
    shutdown_event: asyncio.Event | None = None,
) -> bool:
    """Block until all dependent resources are ready or shutdown is requested.

    Args:
        resources: The resources to wait for.
        poll_interval: The interval to poll for resource status.
        timeout: The timeout for the wait operation.

    Returns:
        True if all resources are ready, False if timeout or shutdown.

    Raises:
        CancelledError: If the wait operation is cancelled.
        TimeoutError: If the wait operation times out.
    """

    resources = resources if isinstance(resources, list) else [resources]
    deadline = asyncio.get_event_loop().time() + timeout
    while True:
        if shutdown_event and shutdown_event.is_set():
            return False
        if all(r.is_ready() for r in resources):
            return True
        if asyncio.get_event_loop().time() >= deadline:
            return False
        await asyncio.sleep(poll_interval)



import asyncio
import json
import logging
import typing
from contextlib import AsyncExitStack, suppress
from itertools import count
from logging import getLogger
from typing import Any, cast

import aiohttp
import anyio
from aiohttp import ClientConnectorError, ClientOSError, ClientTimeout, ServerDisconnectedError, WSMsgType
from aiohttp.client_exceptions import ClientConnectionResetError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    retry_if_not_exception_type,
    stop_after_attempt,
    wait_exponential_jitter,
)

from hassette.events import HassetteSimpleEvent, create_event_from_hass
from hassette.exceptions import (
    ConnectionClosedError,
    CouldNotFindHomeAssistantError,
    FailedMessageError,
    InvalidAuthError,
    ResourceNotReadyError,
    RetryableConnectionClosedError,
)
from hassette.resources.base import Service
from hassette.types import Topic

if typing.TYPE_CHECKING:
    from hassette import Hassette
    from hassette.events.hass.raw import HassEventEnvelopeDict

LOGGER = getLogger(__name__)

# classify errors once (easy to audit/change later)
NON_RETRYABLE = (InvalidAuthError, asyncio.CancelledError)
RETRYABLE = (
    RetryableConnectionClosedError,
    ServerDisconnectedError,
    ClientConnectorError,
    ClientOSError,
    CouldNotFindHomeAssistantError,
)


class WebsocketService(Service):
    url: str
    """WebSocket URL to connect to."""

    _stack: AsyncExitStack
    """Async context stack for managing resources."""

    _session: aiohttp.ClientSession | None
    """HTTP client session for making requests."""

    _ws: aiohttp.ClientWebSocketResponse | None
    """WebSocket connection."""

    _response_futures: dict[int, asyncio.Future[Any]]
    """Mapping of message IDs to futures for awaiting responses."""

    _seq: typing.Iterator[int]
    """Iterator for generating unique message IDs."""

    _recv_task: asyncio.Task | None
    """Task for receiving messages from the WebSocket."""

    _subscription_ids: set[int]
    """Set of active subscription IDs."""

    _connect_lock: asyncio.Lock
    """Lock to prevent concurrent connection attempts."""

    @classmethod
    def create(cls, hassette: "Hassette"):
        inst = cls(hassette=hassette, parent=hassette)
        inst.url = inst.hassette.ws_url
        inst._stack = AsyncExitStack()
        inst._session = None
        inst._ws = None
        inst._response_futures = {}
        inst._seq = count(1)

        inst._recv_task = None
        inst._subscription_ids = set()
        inst._connect_lock = asyncio.Lock()  # if you don't already have it
        return inst

    @property
    def config_log_level(self):
        """Return the log level from the config for this resource."""
        return self.hassette.config.websocket_log_level

    @property
    def resp_timeout_seconds(self) -> int:
        return self.hassette.config.websocket_response_timeout_seconds

    @property
    def connection_timeout_seconds(self) -> int:
        return self.hassette.config.websocket_connection_timeout_seconds

    @property
    def total_timeout_seconds(self) -> int:
        return self.hassette.config.websocket_total_timeout_seconds

    @property
    def heartbeat_interval_seconds(self) -> int:
        return self.hassette.config.websocket_heartbeat_interval_seconds

    @property
    def authentication_timeout_seconds(self) -> int:
        return self.hassette.config.websocket_authentication_timeout_seconds

    @property
    def connected(self) -> bool:
        if self._ws is None:
            return False

        if self._ws._conn is None:
            return False

        return not self._ws._conn.closed

    def get_next_message_id(self) -> int:
        """Get the next message ID."""
        return next(self._seq)

    async def before_shutdown(self) -> None:
        await self._send_connection_lost_event()

    async def serve(self) -> None:
        """Connect to the WebSocket and run the receive loop."""
        async with self._connect_lock:
            timeout = ClientTimeout(connect=self.connection_timeout_seconds, total=self.total_timeout_seconds)

            async with aiohttp.ClientSession(timeout=timeout) as session:
                self._recv_task = await self._make_connection(session)

                # Keep running until recv loop ends (disconnect, error, etc.)
                await self._recv_task

    async def _make_connection(self, session: aiohttp.ClientSession) -> asyncio.Task:
        # inner function so we can use `self` in the retry decorator
        @retry(
            retry=retry_if_not_exception_type(NON_RETRYABLE) | retry_if_exception_type(RETRYABLE),
            wait=wait_exponential_jitter(initial=1, max=32),
            stop=stop_after_attempt(5),
            reraise=True,
            before_sleep=before_sleep_log(self.logger, logging.WARNING),
        )
        async def _inner_connect():
            self._session = session

            try:
                self._ws = await session.ws_connect(
                    self.url, heartbeat=self.heartbeat_interval_seconds, ssl=self.hassette.config.verify_ssl
                )
            except ClientConnectorError as exc:
                if exc.__cause__ and isinstance(exc.__cause__, ConnectionRefusedError):
                    raise CouldNotFindHomeAssistantError(self.url) from exc.__cause__
                raise

            self.logger.debug("Connected to WebSocket at %s", self.url)
            await self.authenticate()

            # mark ready before subscribing, otherwise we'll raise an exception due to not ready status
            self.mark_ready(reason="WebSocket connected and authenticated")

            # start reader first so send_and_wait can get replies
            recv_task = self.task_bucket.spawn(self._recv_loop(), name="ws:recv")

            await self._send_connection_established_event()
            # subscribe to events
            self._subscription_ids.add(await self._subscribe_events())
            return recv_task

        return await _inner_connect()

    async def _recv_loop(self) -> None:
        while True:
            await self._raw_recv()

    async def _subscribe_events(self, event_type: str | None = None) -> int:
        """Subscribe to HA events; returns the subscription ID (the message id you sent)."""
        payload: dict[str, Any] = {"type": "subscribe_events"}
        if event_type is not None:
            payload["event_type"] = event_type  # omit to get all events

        payload["id"] = sub_id = self.get_next_message_id()
        # Use send_and_wait so we see success/error deterministically
        await self.send_and_wait(**payload)
        # HA replies with {'id': <same>, 'type': 'result', 'success': True}
        # We return our own id as the subscription handle for unsubscribe
        return sub_id

    async def cleanup(self) -> None:
        """Cleanup resources after the WebSocket connection is closed."""

        # Set exceptions for all pending response futures
        for fut in list(self._response_futures.values()):
            if not fut.done():
                fut.set_exception(RetryableConnectionClosedError("WebSocket disconnected"))
        self._response_futures.clear()

        # Try to unsubscribe (best-effort; ignore errors if socket is going away)
        if self._ws and not self._ws.closed and self._subscription_ids:
            for sid in list(self._subscription_ids):
                with suppress(Exception):
                    await self.send_json(type="unsubscribe_events", subscription=sid)
            self._subscription_ids.clear()

        # Stop the recv loop
        if self._recv_task:
            self._recv_task.cancel()
            await asyncio.gather(self._recv_task, return_exceptions=True)
            self._recv_task = None

        # Close the WebSocket
        if self._ws and not self._ws.closed:
            await self._ws.close(
                code=aiohttp.WSCloseCode.GOING_AWAY,
                message=b"Shutting down WebSocket connection",
            )
            self.logger.debug("Closed WebSocket with code %s", aiohttp.WSCloseCode.GOING_AWAY)

        # Close the aiohttp session
        if self._session:
            await self._session.close()
            self.logger.debug("Closed aiohttp session")

        await super().cleanup()

    async def send_and_wait(self, **data: Any) -> dict[str, Any]:
        """Send a message and wait for a response.

        Args:
            **data: The data to send as a JSON payload.

        Returns:
            The response data from the WebSocket.

        Raises:
            FailedMessageError: If sending the message fails or times out.
        """

        if "id" not in data:
            data["id"] = msg_id = self.get_next_message_id()
        else:
            msg_id = data["id"]

        fut = self.hassette.loop.create_future()
        self._response_futures[msg_id] = fut
        try:
            await self.send_json(**data)
            return await asyncio.wait_for(fut, timeout=self.resp_timeout_seconds)
        except TimeoutError:
            raise FailedMessageError(f"Response timed out after {self.resp_timeout_seconds}s (data: {data})") from None
        finally:
            self._response_futures.pop(msg_id, None)

    def _respond_if_necessary(self, message: dict) -> None:
        if message.get("type") != "result":
            return

        msg_id = message.get("id")

        if not msg_id:
            self.logger.warning("Received result message without ID: %s", message)
            return

        fut = self._response_futures.get(msg_id)
        if not fut or fut.done():
            return

        if message.get("success"):
            fut.set_result(message.get("result"))

        else:
            err = (message.get("error") or {}).get("message", "Unknown error")
            fut.set_exception(FailedMessageError.from_error_response(err, original_data=message))

    async def send_json(self, **data) -> None:
        """Send a JSON payload over the WebSocket connection, with an incrementing message ID.

        Args:
            **data: The data to send as a JSON payload.

        Raises:
            FailedMessageError: If sending the message fails.
        """

        if not self.ready_event.is_set():
            raise ResourceNotReadyError("WebSocket is not ready")

        self.logger.debug("Sending WebSocket message: %s", data)

        if not isinstance(data, dict):
            raise TypeError("Payload must be a dictionary, got %s", type(data).__name__)

        if not self.connected:
            raise ConnectionClosedError("WebSocket connection is not established")

        # this should never be an issue because self.connected checks for this already
        assert self._ws is not None, "WebSocket must be initialized before sending messages"

        if "id" not in data:
            data["id"] = self.get_next_message_id()

        try:
            await self._ws.send_json(data)
        except ClientConnectionResetError:
            self.logger.error("WebSocket connection reset by peer")
            raise

        except Exception as e:
            self.logger.exception("Exception when sending message: %s", data)
            raise FailedMessageError(f"Failed to send message: {data}") from e

    async def authenticate(self) -> None:
        """Authenticate with the Home Assistant WebSocket API."""

        assert self._ws, "WebSocket must be initialized before authenticating"
        token = self.hassette.config.token
        truncated_token = self.hassette.config.truncated_token
        ws_url = self.hassette.ws_url

        with anyio.fail_after(self.authentication_timeout_seconds):
            msg = await self._ws.receive_json()
            assert msg["type"] == "auth_required"
            await self._ws.send_json({"type": "auth", "access_token": token})
            msg = await self._ws.receive_json()

            # happy path
            if msg["type"] == "auth_ok":
                self.logger.debug("Authenticated successfully with Home Assistant at %s", ws_url)
                return

            if msg["type"] == "auth_invalid":
                self.logger.critical(
                    "Invalid authentication (using token %s) for Home Assistant instance at %s",
                    truncated_token,
                    ws_url,
                )
                raise InvalidAuthError(f"Authentication failed - invalid access token ({truncated_token}) for {ws_url}")

            raise RuntimeError(f"Unexpected authentication response: {msg}")

    async def _raw_recv(self) -> None:
        """Receive a raw WebSocket frame.

        Raises:
            ConnectionClosedError: If the connection is closed.
        """

        if not self._ws:
            raise RuntimeError("WebSocket connection is not established")

        if self._ws.closed:
            raise RetryableConnectionClosedError("WebSocket connection is closed")

        msg = await self._ws.receive()
        msg_type, raw = msg.type, msg.data

        if msg_type == WSMsgType.TEXT:
            try:
                data = json.loads(raw) if raw else {}
            except json.JSONDecodeError:
                self.logger.exception("Invalid JSON received: %s", raw)
                return

            await self._dispatch(data)
            return

        if msg_type == WSMsgType.BINARY:
            self.logger.warning("Received binary message, which is not expected: %r", raw)
            return

        if msg_type in {WSMsgType.CLOSE, WSMsgType.CLOSED}:
            raise RetryableConnectionClosedError(f"WebSocket closed by peer ({msg_type!r})")

        # took a while to track this one down - we need to cancel if we get told that the connection is closing
        if msg_type == WSMsgType.CLOSING:
            self.logger.debug("WebSocket is closing - exiting receive loop")
            raise RetryableConnectionClosedError("WebSocket is closing")

        self.logger.warning("Received unexpected message type: %r", msg_type)

    async def _dispatch(self, data: dict[str, Any]) -> None:
        try:
            match data.get("type"):
                case "event":
                    await self._dispatch_hass_event(cast("HassEventEnvelopeDict", data))
                case "result":
                    self._respond_if_necessary(data)
                case other:
                    self.logger.debug("Ignoring unknown message type: %s", other)
        except Exception:
            self.logger.exception("Failed to dispatch message: %s", data)

    async def _dispatch_hass_event(self, data: "HassEventEnvelopeDict") -> None:
        """Dispatch a Home Assistant event to the event bus."""
        event = create_event_from_hass(data)
        await self.hassette.send_event(event.topic, event)

    async def _send_connection_lost_event(self) -> None:
        """Send a connection lost event to the event bus."""
        event = HassetteSimpleEvent.create_event(topic=Topic.HASSETTE_EVENT_WEBSOCKET_DISCONNECTED)
        await self.hassette.send_event(event.topic, event)

    async def _send_connection_established_event(self) -> None:
        """Send a connection established event to the event bus."""
        event = HassetteSimpleEvent.create_event(topic=Topic.HASSETTE_EVENT_WEBSOCKET_CONNECTED)
        await self.hassette.send_event(event.topic, event)



from typing import Any, Literal

from pydantic import Field
from whenever import Instant, ZonedDateTime

from hassette.utils.date_utils import convert_utc_timestamp_to_system_tz

from .base import AttributesBase, BoolBaseState, DateTimeBaseState, NumericBaseState, StringBaseState


class InputAttributesBase(AttributesBase):
    """Base attributes class for all input states."""

    editable: bool | None = Field(default=None)


class InputBooleanState(BoolBaseState):
    """Representation of a Home Assistant input_boolean state.

    See: https://www.home-assistant.io/integrations/input_boolean/
    """

    domain: Literal["input_boolean"]

    attributes: InputAttributesBase


class InputButtonState(DateTimeBaseState):
    """Representation of a Home Assistant input_button state.

    See: https://www.home-assistant.io/integrations/input_button/
    """

    domain: Literal["input_button"]

    attributes: InputAttributesBase


class InputDatetimeAttributes(InputAttributesBase):
    has_date: bool | None = Field(default=None)
    has_time: bool | None = Field(default=None)
    year: int | None = Field(default=None)
    month: int | None = Field(default=None)
    day: int | None = Field(default=None)
    hour: int | None = Field(default=None)
    minute: int | None = Field(default=None)
    second: int | None = Field(default=None)
    timestamp: float | None = Field(default=None)

    @property
    def timestamp_as_instant(self) -> Instant | None:
        if self.timestamp is None:
            return None
        return Instant.from_timestamp(self.timestamp)

    @property
    def timestamp_as_system_datetime(self) -> ZonedDateTime | None:
        if self.timestamp is None:
            return None
        return convert_utc_timestamp_to_system_tz(self.timestamp)


class InputDatetimeState(DateTimeBaseState):
    """Representation of a Home Assistant input_datetime state.

    See: https://www.home-assistant.io/integrations/input_datetime/
    """

    domain: Literal["input_datetime"]

    attributes: InputDatetimeAttributes


class InputNumberAttributes(InputAttributesBase):
    max: float | None = Field(default=None)
    initial: float | None = Field(default=None)
    step: int | float | None = Field(default=None)
    mode: str | None = Field(default=None)
    min: int | float | None = Field(default=None)


class InputNumberState(NumericBaseState):
    """Representation of a Home Assistant input_number state.

    See: https://www.home-assistant.io/integrations/input_number/
    """

    domain: Literal["input_number"]

    attributes: InputNumberAttributes


class InputSelectAttributes(InputAttributesBase):
    options: list[str] = Field(default_factory=list)


class InputSelectState(StringBaseState):
    """Representation of a Home Assistant input_select state.

    See: https://www.home-assistant.io/integrations/input_select/
    """

    domain: Literal["input_select"]

    attributes: InputSelectAttributes


class InputTextAttributes(InputAttributesBase):
    min: int | float | None = Field(default=None)
    max: int | float | None = Field(default=None)
    pattern: Any | None = Field(default=None)
    mode: str | None = Field(default=None)


class InputTextState(StringBaseState):
    """Representation of a Home Assistant input_text state.

    See: https://www.home-assistant.io/integrations/input_text/
    """

    domain: Literal["input_text"]

    attributes: InputTextAttributes



import logging
import sys
import threading
from contextlib import suppress
from typing import Literal

import coloredlogs

FORMAT_DATE = "%Y-%m-%d"
FORMAT_TIME = "%H:%M:%S"
FORMAT_DATETIME = f"{FORMAT_DATE} {FORMAT_TIME}"
FMT = "%(asctime)s.%(msecs)03d %(levelname)s %(name)s.%(funcName)s:%(lineno)d ─ %(message)s"

# TODO: remove coloredlogs and roll our own? or use colorlogs
# coloredlogs is unmaintained and parts of it are broken on Python >3.13


def enable_logging(
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
) -> None:
    """Set up the logging"""

    logger = logging.getLogger("hassette")

    # Set the base hassette logger
    logger.setLevel(log_level)

    # don't propagate to root - if someone wants to do a basicConfig on root we don't want
    # our logs going there too.
    logger.propagate = False

    # Clear any old handlers
    logger.handlers.clear()

    # NOTSET - don't clamp child logs
    # this is the kicker - if the handler is filtered then it doesn't matter what we set the
    # logger to, it won't log anything lower than the handler's level.
    # So we set the handler to NOTSET and clamp the logger itself.
    # don't know why it took me five years to learn this.
    coloredlogs.install(level=logging.NOTSET, logger=logger, fmt=FMT, datefmt=FORMAT_DATETIME)

    # reset hassette logger to desired level, as coloredlogs.install sets it to WARNING
    logger.setLevel(log_level)

    # coloredlogs does something funky to the root logger and i can't figure out what
    # so for now i'm just resorting to this
    with suppress(IndexError):
        logging.getLogger().handlers.pop(0)

    # here and below were pulled from Home Assistant

    # Capture warnings.warn(...) and friends messages in logs.
    # The standard destination for them is stderr, which may end up unnoticed.
    # This way they're where other messages are, and can be filtered as usual.
    logging.captureWarnings(True)

    # Suppress overly verbose logs from libraries that aren't helpful
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("aiohttp.access").setLevel(logging.WARNING)
    logging.getLogger("httpx").setLevel(logging.WARNING)

    sys.excepthook = lambda *args: logging.getLogger().exception("Uncaught exception", exc_info=args)
    threading.excepthook = lambda args: logging.getLogger().exception(
        "Uncaught thread exception",
        exc_info=(args.exc_type, args.exc_value, args.exc_traceback),  # pyright: ignore[reportArgumentType]
    )



from typing import Any, Literal

from pydantic import Field

from .base import AttributesBase, StringBaseState


class TextAttributes(AttributesBase):
    editable: bool | None = Field(default=None)
    min: int | float | None = Field(default=None)
    max: int | float | None = Field(default=None)
    pattern: Any | None = Field(default=None)
    mode: str | None = Field(default=None)


class TextState(StringBaseState):
    """Representation of a Home Assistant text state.

    See: https://www.home-assistant.io/integrations/text/
    """

    domain: Literal["text"]

    attributes: TextAttributes
